/* A Bison parser, made by GNU Bison 3.0.2.  */

/* Bison implementation for Yacc-like parsers in C

   Copyright (C) 1984, 1989-1990, 2000-2013 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

/* As a special exception, you may create a larger work that contains
   part or all of the Bison parser skeleton and distribute that work
   under terms of your choice, so long as that work isn't itself a
   parser generator using the skeleton or a modified version thereof
   as a parser skeleton.  Alternatively, if you modify or redistribute
   the parser skeleton itself, you may (at your option) remove this
   special exception, which will cause the skeleton and the resulting
   Bison output files to be licensed under the GNU General Public
   License without this special exception.

   This special exception was added by the Free Software Foundation in
   version 2.2 of Bison.  */

/* C LALR(1) parser skeleton written by Richard Stallman, by
   simplifying the original so-called "semantic" parser.  */

/* All symbols defined below should begin with yy or YY, to avoid
   infringing on user name space.  This should be done even for local
   variables, as they might otherwise be expanded by user macros.
   There are some unavoidable exceptions within include files to
   define necessary library symbols; they are noted "INFRINGES ON
   USER NAME SPACE" below.  */

/* Identify Bison output.  */
#define YYBISON 1

/* Bison version.  */
#define YYBISON_VERSION "3.0.2"

/* Skeleton name.  */
#define YYSKELETON_NAME "yacc.c"

/* Pure parsers.  */
#define YYPURE 0

/* Push parsers.  */
#define YYPUSH 0

/* Pull parsers.  */
#define YYPULL 1


/* Substitute the variable and function names.  */
#define yyparse         PLparse
#define yylex           PLlex
#define yyerror         PLerror
#define yydebug         PLdebug
#define yynerrs         PLnerrs

#define yylval          PLlval
#define yychar          PLchar

/* Copy the first part of user declarations.  */
#line 1 "PL.y" /* yacc.c:339  */

/*****************************************************************************/
/*!
 * \file PL.y
 * 
 * Author: Sergey Berezin
 * 
 * Created: Feb 06 03:00:43 GMT 2003
 *
 * <hr>
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/
/* PL.y
   Aaron Stump, 4/14/01

   This file contains the bison code for the parser that reads in CVC
   commands in the presentation language (hence "PL").
*/

#include "vc.h"
#include "parser_exception.h"
#include "parser_temp.h"
#include "rational.h"

// Exported shared data
namespace CVC3 {
  extern ParserTemp* parserTemp;
}
// Define shortcuts for various things
#define TMP CVC3::parserTemp
#define EXPR CVC3::parserTemp->expr
#define VC (CVC3::parserTemp->vc)
#define RAT(args) CVC3::newRational args

// Suppress the bogus warning suppression in bison (it generates
// compile error)
#undef __GNUC_MINOR__

/* stuff that lives in PL.lex */
extern int PLlex(void);

int PLerror(const char *s)
{
  std::ostringstream ss;
  ss << CVC3::parserTemp->fileName << ":" << CVC3::parserTemp->lineNum
     << ": " << s;
  return CVC3::parserTemp->error(ss.str());
}


namespace CVC3 {

// File-static auxiliary funcion to wrap accessors upto the index i
// (not inclusive) around "e".  Used to rebuild the UPDATE expressions.
static Expr wrapAccessors(const Expr& e,
			  const std::vector<Expr>& accessors,
			  const int i) {
  DebugAssert((int)accessors.size() >= i, "wrapAccessors: too few accessors");
  Expr res(e);
  for(int j=0; j<i; j++) {
    const Expr& acc = accessors[j];
    switch(acc.getKind()) {
    case ID: // Datatype: apply the accessor directly to 'res'
      res = VC->listExpr(acc, res);
      break;
    case RAW_LIST:
      DebugAssert(acc.arity() == 2 && acc[0].getKind() == ID,
		  "PL.y: wrong number of arguments in the accessor of WITH: "
		  +acc.toString());
      res = VC->listExpr(acc[0], res, acc[1]);
      break;
    default:
      DebugAssert(false, "PL.y: Bad accessor in WITH: "+acc.toString());
    }
  }
  return res;
}

// Convert a complex WITH statement into a bunch of nested simple UPDATEs

// Updator "e1 WITH accessors := e2" is represented as
// (e1 (accessor1 accessor2 ... ) e2), where the accessors are:
// (READ idx)
// ID (datatype constructor's argument)
// (RECORD_SELECT ID)
// and (TUPLE_SELECT num).
 
// We rebuild this into nested RECORD_UPDATE, WRITE,
// TUPLE_UPDATE, and DATATYPE_UPDATE expressions.  For example:
//
// e WITH [idx] car.f := newVal  is transformed into
// e WITH [idx] := (e[idx] WITH car := (car(e[idx]) WITH .f := newVal))
// which is represented as
// (WRITE e idx
//        (DATATYPE_UPDATE (READ e idx) car
//                         (RECORD_UPDATE (car (READ e idx)) f newVal))).
// Here "car" and "f" are identifiers (ID "car") and (ID "f").

static Expr PLprocessUpdate(const CVC3::Expr& e,
			    const CVC3::Expr& update) {
  TRACE("parser verbose", "PLprocessUpdate(", e, ") {");
  DebugAssert(update.arity() == 2,"PL.y: Bad WITH statement: "
	      +update.toString());
  // Rebuild accessors: resolve ID in (READ (ID "name")) and leave
  // the rest alone
  TRACE("parser verbose", "PLprocessUpdate: update[0] = ", update[0], "");
  std::vector<Expr> acc;
  for(Expr::iterator i=update[0].begin(), iend=update[0].end(); i!=iend; ++i) {
    TRACE("parser verbose", "*i = ", *i, "");
    acc.push_back(*i);
  }
  TRACE("parser verbose", "acc.size() = ", acc.size(), "");
  TRACE("parser verbose", "accessors = ", VC->listExpr(acc), "");
  // 'res' serves as the accumulator of updators; initially it is
  // newVal (which is update[1]).  We run through accessors in reverse
  // order, wrapping this accumulator with the corresponding
  // updators.
  Expr res(update[1]);
  // Rebuilding the original expression "e"
  // The main loop
  for(int i=acc.size()-1; i>=0; i--) {
    Expr ac(acc[i]);  // The current accessor
    TRACE("parser verbose", "ac["+int2string(i)+"] = ", ac, "");
    // "e" with all preceding accessors
    Expr tmp(wrapAccessors(e, acc, i));
    TRACE("parser verbose", "tmp = ", tmp, "");
    switch(ac.getKind()) {
    case ID: // Datatype update
      res = VC->listExpr("_DATATYPE_UPDATE", tmp, ac, res);
      break;
    case RAW_LIST: {
      const std::string& kind = ac[0][0].getString();
      if(kind == "_READ") // Array update
	res = VC->listExpr("_WRITE", tmp, ac[1], res);
      else if(kind == "_RECORD_SELECT") // Record update
	res = VC->listExpr("_RECORD_UPDATE", tmp, ac[1], res);
      else if(kind == "_TUPLE_SELECT") // Tuple element
	res = VC->listExpr("_TUPLE_UPDATE", tmp, ac[1], res);
      else
	DebugAssert(false, "PL.y: Bad accessor in WITH: "+ac.toString());
      break;
    }
    default:
      DebugAssert(false, "PL.y: Bad accessor in WITH: "+ac.toString());
    }
  }
  TRACE("parser verbose", "PLprocessUpdate => ", res, " }");
  return res;
}


// Process a vector of simultaneous updates (from a single WITH)
static Expr PLprocessUpdates(const CVC3::Expr& e,
			     const std::vector<CVC3::Expr>& updates,
			     size_t idx=0) {
  // Processed all the updates
  if(idx >= updates.size()) return e;
  return PLprocessUpdates(PLprocessUpdate(e, updates[idx]), updates, idx+1);
}



} // end of namespace CVC3

#define YYLTYPE_IS_TRIVIAL 1
#define YYMAXDEPTH 10485760
/* Prefer YYERROR_VERBOSE over %error-verbose to avoid errors in older bison */
#define YYERROR_VERBOSE 1

#line 252 "parsePL.cpp" /* yacc.c:339  */

# ifndef YY_NULLPTR
#  if defined __cplusplus && 201103L <= __cplusplus
#   define YY_NULLPTR nullptr
#  else
#   define YY_NULLPTR 0
#  endif
# endif

/* Enabling verbose error messages.  */
#ifdef YYERROR_VERBOSE
# undef YYERROR_VERBOSE
# define YYERROR_VERBOSE 1
#else
# define YYERROR_VERBOSE 0
#endif

/* In a future release of Bison, this section will be replaced
   by #include "parsePL.hpp".  */
#ifndef YY_PL_PARSEPL_HPP_INCLUDED
# define YY_PL_PARSEPL_HPP_INCLUDED
/* Debug traces.  */
#ifndef YYDEBUG
# define YYDEBUG 1
#endif
#if YYDEBUG
extern int PLdebug;
#endif

/* Token type.  */
#ifndef YYTOKENTYPE
# define YYTOKENTYPE
  enum yytokentype
  {
    DISTINCT_TOK = 258,
    AND_TOK = 259,
    ARRAY_TOK = 260,
    BOOLEAN_TOK = 261,
    DATATYPE_TOK = 262,
    ELSE_TOK = 263,
    ELSIF_TOK = 264,
    END_TOK = 265,
    ENDIF_TOK = 266,
    EXISTS_TOK = 267,
    FALSELIT_TOK = 268,
    FORALL_TOK = 269,
    IN_TOK = 270,
    IF_TOK = 271,
    LAMBDA_TOK = 272,
    SIMULATE_TOK = 273,
    LET_TOK = 274,
    NOT_TOK = 275,
    IS_INTEGER_TOK = 276,
    OF_TOK = 277,
    OR_TOK = 278,
    REAL_TOK = 279,
    INT_TOK = 280,
    SUBTYPE_TOK = 281,
    BITVECTOR_TOK = 282,
    THEN_TOK = 283,
    TRUELIT_TOK = 284,
    TYPE_TOK = 285,
    WITH_TOK = 286,
    XOR_TOK = 287,
    TCC_TOK = 288,
    PATTERN_TOK = 289,
    DONE_TOK = 290,
    DOTDOT_TOK = 291,
    ASSIGN_TOK = 292,
    NEQ_TOK = 293,
    IMPLIES_TOK = 294,
    IFF_TOK = 295,
    PLUS_TOK = 296,
    MINUS_TOK = 297,
    MULT_TOK = 298,
    POW_TOK = 299,
    DIV_TOK = 300,
    MOD_TOK = 301,
    INTDIV_TOK = 302,
    LT_TOK = 303,
    LE_TOK = 304,
    GT_TOK = 305,
    GE_TOK = 306,
    FLOOR_TOK = 307,
    LEFTSHIFT_TOK = 308,
    RIGHTSHIFT_TOK = 309,
    BVPLUS_TOK = 310,
    BVSUB_TOK = 311,
    BVUDIV_TOK = 312,
    BVSDIV_TOK = 313,
    BVUREM_TOK = 314,
    BVSREM_TOK = 315,
    BVSMOD_TOK = 316,
    BVSHL_TOK = 317,
    BVASHR_TOK = 318,
    BVLSHR_TOK = 319,
    BVUMINUS_TOK = 320,
    BVMULT_TOK = 321,
    SQHASH_TOK = 322,
    HASHSQ_TOK = 323,
    PARENHASH_TOK = 324,
    HASHPAREN_TOK = 325,
    ARROW_TOK = 326,
    BVNEG_TOK = 327,
    BVAND_TOK = 328,
    MID_TOK = 329,
    BVXOR_TOK = 330,
    BVNAND_TOK = 331,
    BVNOR_TOK = 332,
    BVCOMP_TOK = 333,
    BVXNOR_TOK = 334,
    CONCAT_TOK = 335,
    BVTOINT_TOK = 336,
    INTTOBV_TOK = 337,
    BOOLEXTRACT_TOK = 338,
    BVLT_TOK = 339,
    BVGT_TOK = 340,
    BVLE_TOK = 341,
    BVGE_TOK = 342,
    SX_TOK = 343,
    BVZEROEXTEND_TOK = 344,
    BVREPEAT_TOK = 345,
    BVROTL_TOK = 346,
    BVROTR_TOK = 347,
    BVSLT_TOK = 348,
    BVSGT_TOK = 349,
    BVSLE_TOK = 350,
    BVSGE_TOK = 351,
    ARITH_VAR_ORDER_TOK = 352,
    ASSERT_TOK = 353,
    QUERY_TOK = 354,
    CHECKSAT_TOK = 355,
    CONTINUE_TOK = 356,
    RESTART_TOK = 357,
    HELP_TOK = 358,
    DBG_TOK = 359,
    TRACE_TOK = 360,
    UNTRACE_TOK = 361,
    OPTION_TOK = 362,
    TRANSFORM_TOK = 363,
    PRINT_TOK = 364,
    PRINT_TYPE_TOK = 365,
    CALL_TOK = 366,
    ECHO_TOK = 367,
    INCLUDE_TOK = 368,
    DUMP_PROOF_TOK = 369,
    DUMP_ASSUMPTIONS_TOK = 370,
    DUMP_SIG_TOK = 371,
    DUMP_TCC_TOK = 372,
    DUMP_TCC_ASSUMPTIONS_TOK = 373,
    DUMP_TCC_PROOF_TOK = 374,
    DUMP_CLOSURE_TOK = 375,
    DUMP_CLOSURE_PROOF_TOK = 376,
    WHERE_TOK = 377,
    ASSERTIONS_TOK = 378,
    ASSUMPTIONS_TOK = 379,
    COUNTEREXAMPLE_TOK = 380,
    COUNTERMODEL_TOK = 381,
    PUSH_TOK = 382,
    POP_TOK = 383,
    POPTO_TOK = 384,
    PUSH_SCOPE_TOK = 385,
    POP_SCOPE_TOK = 386,
    POPTO_SCOPE_TOK = 387,
    RESET_TOK = 388,
    CONTEXT_TOK = 389,
    FORGET_TOK = 390,
    GET_TYPE_TOK = 391,
    CHECK_TYPE_TOK = 392,
    GET_CHILD_TOK = 393,
    GET_OP_TOK = 394,
    SUBSTITUTE_TOK = 395,
    UMINUS_TOK = 396,
    ID_TOK = 397,
    STRINGLIT_TOK = 398,
    NUMERAL_TOK = 399,
    HEX_TOK = 400,
    BINARY_TOK = 401
  };
#endif
/* Tokens.  */
#define DISTINCT_TOK 258
#define AND_TOK 259
#define ARRAY_TOK 260
#define BOOLEAN_TOK 261
#define DATATYPE_TOK 262
#define ELSE_TOK 263
#define ELSIF_TOK 264
#define END_TOK 265
#define ENDIF_TOK 266
#define EXISTS_TOK 267
#define FALSELIT_TOK 268
#define FORALL_TOK 269
#define IN_TOK 270
#define IF_TOK 271
#define LAMBDA_TOK 272
#define SIMULATE_TOK 273
#define LET_TOK 274
#define NOT_TOK 275
#define IS_INTEGER_TOK 276
#define OF_TOK 277
#define OR_TOK 278
#define REAL_TOK 279
#define INT_TOK 280
#define SUBTYPE_TOK 281
#define BITVECTOR_TOK 282
#define THEN_TOK 283
#define TRUELIT_TOK 284
#define TYPE_TOK 285
#define WITH_TOK 286
#define XOR_TOK 287
#define TCC_TOK 288
#define PATTERN_TOK 289
#define DONE_TOK 290
#define DOTDOT_TOK 291
#define ASSIGN_TOK 292
#define NEQ_TOK 293
#define IMPLIES_TOK 294
#define IFF_TOK 295
#define PLUS_TOK 296
#define MINUS_TOK 297
#define MULT_TOK 298
#define POW_TOK 299
#define DIV_TOK 300
#define MOD_TOK 301
#define INTDIV_TOK 302
#define LT_TOK 303
#define LE_TOK 304
#define GT_TOK 305
#define GE_TOK 306
#define FLOOR_TOK 307
#define LEFTSHIFT_TOK 308
#define RIGHTSHIFT_TOK 309
#define BVPLUS_TOK 310
#define BVSUB_TOK 311
#define BVUDIV_TOK 312
#define BVSDIV_TOK 313
#define BVUREM_TOK 314
#define BVSREM_TOK 315
#define BVSMOD_TOK 316
#define BVSHL_TOK 317
#define BVASHR_TOK 318
#define BVLSHR_TOK 319
#define BVUMINUS_TOK 320
#define BVMULT_TOK 321
#define SQHASH_TOK 322
#define HASHSQ_TOK 323
#define PARENHASH_TOK 324
#define HASHPAREN_TOK 325
#define ARROW_TOK 326
#define BVNEG_TOK 327
#define BVAND_TOK 328
#define MID_TOK 329
#define BVXOR_TOK 330
#define BVNAND_TOK 331
#define BVNOR_TOK 332
#define BVCOMP_TOK 333
#define BVXNOR_TOK 334
#define CONCAT_TOK 335
#define BVTOINT_TOK 336
#define INTTOBV_TOK 337
#define BOOLEXTRACT_TOK 338
#define BVLT_TOK 339
#define BVGT_TOK 340
#define BVLE_TOK 341
#define BVGE_TOK 342
#define SX_TOK 343
#define BVZEROEXTEND_TOK 344
#define BVREPEAT_TOK 345
#define BVROTL_TOK 346
#define BVROTR_TOK 347
#define BVSLT_TOK 348
#define BVSGT_TOK 349
#define BVSLE_TOK 350
#define BVSGE_TOK 351
#define ARITH_VAR_ORDER_TOK 352
#define ASSERT_TOK 353
#define QUERY_TOK 354
#define CHECKSAT_TOK 355
#define CONTINUE_TOK 356
#define RESTART_TOK 357
#define HELP_TOK 358
#define DBG_TOK 359
#define TRACE_TOK 360
#define UNTRACE_TOK 361
#define OPTION_TOK 362
#define TRANSFORM_TOK 363
#define PRINT_TOK 364
#define PRINT_TYPE_TOK 365
#define CALL_TOK 366
#define ECHO_TOK 367
#define INCLUDE_TOK 368
#define DUMP_PROOF_TOK 369
#define DUMP_ASSUMPTIONS_TOK 370
#define DUMP_SIG_TOK 371
#define DUMP_TCC_TOK 372
#define DUMP_TCC_ASSUMPTIONS_TOK 373
#define DUMP_TCC_PROOF_TOK 374
#define DUMP_CLOSURE_TOK 375
#define DUMP_CLOSURE_PROOF_TOK 376
#define WHERE_TOK 377
#define ASSERTIONS_TOK 378
#define ASSUMPTIONS_TOK 379
#define COUNTEREXAMPLE_TOK 380
#define COUNTERMODEL_TOK 381
#define PUSH_TOK 382
#define POP_TOK 383
#define POPTO_TOK 384
#define PUSH_SCOPE_TOK 385
#define POP_SCOPE_TOK 386
#define POPTO_SCOPE_TOK 387
#define RESET_TOK 388
#define CONTEXT_TOK 389
#define FORGET_TOK 390
#define GET_TYPE_TOK 391
#define CHECK_TYPE_TOK 392
#define GET_CHILD_TOK 393
#define GET_OP_TOK 394
#define SUBSTITUTE_TOK 395
#define UMINUS_TOK 396
#define ID_TOK 397
#define STRINGLIT_TOK 398
#define NUMERAL_TOK 399
#define HEX_TOK 400
#define BINARY_TOK 401

/* Value type.  */
#if ! defined YYSTYPE && ! defined YYSTYPE_IS_DECLARED
typedef union YYSTYPE YYSTYPE;
union YYSTYPE
{
#line 179 "PL.y" /* yacc.c:355  */

  std::string *str;
  CVC3::Expr *node;
  std::vector<CVC3::Expr> *vec;
  int kind;

#line 591 "parsePL.cpp" /* yacc.c:355  */
};
# define YYSTYPE_IS_TRIVIAL 1
# define YYSTYPE_IS_DECLARED 1
#endif


extern YYSTYPE PLlval;

int PLparse (void);

#endif /* !YY_PL_PARSEPL_HPP_INCLUDED  */

/* Copy the second part of user declarations.  */

#line 606 "parsePL.cpp" /* yacc.c:358  */

#ifdef short
# undef short
#endif

#ifdef YYTYPE_UINT8
typedef YYTYPE_UINT8 yytype_uint8;
#else
typedef unsigned char yytype_uint8;
#endif

#ifdef YYTYPE_INT8
typedef YYTYPE_INT8 yytype_int8;
#else
typedef signed char yytype_int8;
#endif

#ifdef YYTYPE_UINT16
typedef YYTYPE_UINT16 yytype_uint16;
#else
typedef unsigned short int yytype_uint16;
#endif

#ifdef YYTYPE_INT16
typedef YYTYPE_INT16 yytype_int16;
#else
typedef short int yytype_int16;
#endif

#ifndef YYSIZE_T
# ifdef __SIZE_TYPE__
#  define YYSIZE_T __SIZE_TYPE__
# elif defined size_t
#  define YYSIZE_T size_t
# elif ! defined YYSIZE_T
#  include <stddef.h> /* INFRINGES ON USER NAME SPACE */
#  define YYSIZE_T size_t
# else
#  define YYSIZE_T unsigned int
# endif
#endif

#define YYSIZE_MAXIMUM ((YYSIZE_T) -1)

#ifndef YY_
# if defined YYENABLE_NLS && YYENABLE_NLS
#  if ENABLE_NLS
#   include <libintl.h> /* INFRINGES ON USER NAME SPACE */
#   define YY_(Msgid) dgettext ("bison-runtime", Msgid)
#  endif
# endif
# ifndef YY_
#  define YY_(Msgid) Msgid
# endif
#endif

#ifndef YY_ATTRIBUTE
# if (defined __GNUC__                                               \
      && (2 < __GNUC__ || (__GNUC__ == 2 && 96 <= __GNUC_MINOR__)))  \
     || defined __SUNPRO_C && 0x5110 <= __SUNPRO_C
#  define YY_ATTRIBUTE(Spec) __attribute__(Spec)
# else
#  define YY_ATTRIBUTE(Spec) /* empty */
# endif
#endif

#ifndef YY_ATTRIBUTE_PURE
# define YY_ATTRIBUTE_PURE   YY_ATTRIBUTE ((__pure__))
#endif

#ifndef YY_ATTRIBUTE_UNUSED
# define YY_ATTRIBUTE_UNUSED YY_ATTRIBUTE ((__unused__))
#endif

#if !defined _Noreturn \
     && (!defined __STDC_VERSION__ || __STDC_VERSION__ < 201112)
# if defined _MSC_VER && 1200 <= _MSC_VER
#  define _Noreturn __declspec (noreturn)
# else
#  define _Noreturn YY_ATTRIBUTE ((__noreturn__))
# endif
#endif

/* Suppress unused-variable warnings by "using" E.  */
#if ! defined lint || defined __GNUC__
# define YYUSE(E) ((void) (E))
#else
# define YYUSE(E) /* empty */
#endif

#if defined __GNUC__ && 407 <= __GNUC__ * 100 + __GNUC_MINOR__
/* Suppress an incorrect diagnostic about yylval being uninitialized.  */
# define YY_IGNORE_MAYBE_UNINITIALIZED_BEGIN \
    _Pragma ("GCC diagnostic push") \
    _Pragma ("GCC diagnostic ignored \"-Wuninitialized\"")\
    _Pragma ("GCC diagnostic ignored \"-Wmaybe-uninitialized\"")
# define YY_IGNORE_MAYBE_UNINITIALIZED_END \
    _Pragma ("GCC diagnostic pop")
#else
# define YY_INITIAL_VALUE(Value) Value
#endif
#ifndef YY_IGNORE_MAYBE_UNINITIALIZED_BEGIN
# define YY_IGNORE_MAYBE_UNINITIALIZED_BEGIN
# define YY_IGNORE_MAYBE_UNINITIALIZED_END
#endif
#ifndef YY_INITIAL_VALUE
# define YY_INITIAL_VALUE(Value) /* Nothing. */
#endif


#if ! defined yyoverflow || YYERROR_VERBOSE

/* The parser invokes alloca or malloc; define the necessary symbols.  */

# ifdef YYSTACK_USE_ALLOCA
#  if YYSTACK_USE_ALLOCA
#   ifdef __GNUC__
#    define YYSTACK_ALLOC __builtin_alloca
#   elif defined __BUILTIN_VA_ARG_INCR
#    include <alloca.h> /* INFRINGES ON USER NAME SPACE */
#   elif defined _AIX
#    define YYSTACK_ALLOC __alloca
#   elif defined _MSC_VER
#    include <malloc.h> /* INFRINGES ON USER NAME SPACE */
#    define alloca _alloca
#   else
#    define YYSTACK_ALLOC alloca
#    if ! defined _ALLOCA_H && ! defined EXIT_SUCCESS
#     include <stdlib.h> /* INFRINGES ON USER NAME SPACE */
      /* Use EXIT_SUCCESS as a witness for stdlib.h.  */
#     ifndef EXIT_SUCCESS
#      define EXIT_SUCCESS 0
#     endif
#    endif
#   endif
#  endif
# endif

# ifdef YYSTACK_ALLOC
   /* Pacify GCC's 'empty if-body' warning.  */
#  define YYSTACK_FREE(Ptr) do { /* empty */; } while (0)
#  ifndef YYSTACK_ALLOC_MAXIMUM
    /* The OS might guarantee only one guard page at the bottom of the stack,
       and a page size can be as small as 4096 bytes.  So we cannot safely
       invoke alloca (N) if N exceeds 4096.  Use a slightly smaller number
       to allow for a few compiler-allocated temporary stack slots.  */
#   define YYSTACK_ALLOC_MAXIMUM 4032 /* reasonable circa 2006 */
#  endif
# else
#  define YYSTACK_ALLOC YYMALLOC
#  define YYSTACK_FREE YYFREE
#  ifndef YYSTACK_ALLOC_MAXIMUM
#   define YYSTACK_ALLOC_MAXIMUM YYSIZE_MAXIMUM
#  endif
#  if (defined __cplusplus && ! defined EXIT_SUCCESS \
       && ! ((defined YYMALLOC || defined malloc) \
             && (defined YYFREE || defined free)))
#   include <stdlib.h> /* INFRINGES ON USER NAME SPACE */
#   ifndef EXIT_SUCCESS
#    define EXIT_SUCCESS 0
#   endif
#  endif
#  ifndef YYMALLOC
#   define YYMALLOC malloc
#   if ! defined malloc && ! defined EXIT_SUCCESS
void *malloc (YYSIZE_T); /* INFRINGES ON USER NAME SPACE */
#   endif
#  endif
#  ifndef YYFREE
#   define YYFREE free
#   if ! defined free && ! defined EXIT_SUCCESS
void free (void *); /* INFRINGES ON USER NAME SPACE */
#   endif
#  endif
# endif
#endif /* ! defined yyoverflow || YYERROR_VERBOSE */


#if (! defined yyoverflow \
     && (! defined __cplusplus \
         || (defined YYSTYPE_IS_TRIVIAL && YYSTYPE_IS_TRIVIAL)))

/* A type that is properly aligned for any stack member.  */
union yyalloc
{
  yytype_int16 yyss_alloc;
  YYSTYPE yyvs_alloc;
};

/* The size of the maximum gap between one aligned stack and the next.  */
# define YYSTACK_GAP_MAXIMUM (sizeof (union yyalloc) - 1)

/* The size of an array large to enough to hold all stacks, each with
   N elements.  */
# define YYSTACK_BYTES(N) \
     ((N) * (sizeof (yytype_int16) + sizeof (YYSTYPE)) \
      + YYSTACK_GAP_MAXIMUM)

# define YYCOPY_NEEDED 1

/* Relocate STACK from its old location to the new one.  The
   local variables YYSIZE and YYSTACKSIZE give the old and new number of
   elements in the stack, and YYPTR gives the new location of the
   stack.  Advance YYPTR to a properly aligned location for the next
   stack.  */
# define YYSTACK_RELOCATE(Stack_alloc, Stack)                           \
    do                                                                  \
      {                                                                 \
        YYSIZE_T yynewbytes;                                            \
        YYCOPY (&yyptr->Stack_alloc, Stack, yysize);                    \
        Stack = &yyptr->Stack_alloc;                                    \
        yynewbytes = yystacksize * sizeof (*Stack) + YYSTACK_GAP_MAXIMUM; \
        yyptr += yynewbytes / sizeof (*yyptr);                          \
      }                                                                 \
    while (0)

#endif

#if defined YYCOPY_NEEDED && YYCOPY_NEEDED
/* Copy COUNT objects from SRC to DST.  The source and destination do
   not overlap.  */
# ifndef YYCOPY
#  if defined __GNUC__ && 1 < __GNUC__
#   define YYCOPY(Dst, Src, Count) \
      __builtin_memcpy (Dst, Src, (Count) * sizeof (*(Src)))
#  else
#   define YYCOPY(Dst, Src, Count)              \
      do                                        \
        {                                       \
          YYSIZE_T yyi;                         \
          for (yyi = 0; yyi < (Count); yyi++)   \
            (Dst)[yyi] = (Src)[yyi];            \
        }                                       \
      while (0)
#  endif
# endif
#endif /* !YYCOPY_NEEDED */

/* YYFINAL -- State number of the termination state.  */
#define YYFINAL  202
/* YYLAST -- Last index in YYTABLE.  */
#define YYLAST   5721

/* YYNTOKENS -- Number of terminals.  */
#define YYNTOKENS  159
/* YYNNTS -- Number of nonterminals.  */
#define YYNNTS  92
/* YYNRULES -- Number of rules.  */
#define YYNRULES  290
/* YYNSTATES -- Number of states.  */
#define YYNSTATES  682

/* YYTRANSLATE[YYX] -- Symbol number corresponding to YYX as returned
   by yylex, with out-of-bounds checking.  */
#define YYUNDEFTOK  2
#define YYMAXUTOK   401

#define YYTRANSLATE(YYX)                                                \
  ((unsigned int) (YYX) <= YYMAXUTOK ? yytranslate[YYX] : YYUNDEFTOK)

/* YYTRANSLATE[TOKEN-NUM] -- Symbol number corresponding to TOKEN-NUM
   as returned by yylex, without out-of-bounds checking.  */
static const yytype_uint8 yytranslate[] =
{
       0,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
     147,   154,     2,     2,   157,     2,   146,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,   155,   153,
       2,   142,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,   144,     2,   156,     2,    35,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,   145,     2,   158,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     2,     2,     2,     2,
       2,     2,     2,     2,     2,     2,     1,     2,     3,     4,
       5,     6,     7,     8,     9,    10,    11,    12,    13,    14,
      15,    16,    17,    18,    19,    20,    21,    22,    23,    24,
      25,    26,    27,    28,    29,    30,    31,    32,    33,    34,
      36,    37,    38,    39,    40,    41,    42,    43,    44,    45,
      46,    47,    48,    49,    50,    51,    52,    53,    54,    55,
      56,    57,    58,    59,    60,    61,    62,    63,    64,    65,
      66,    67,    68,    69,    70,    71,    72,    73,    74,    75,
      76,    77,    78,    79,    80,    81,    82,    83,    84,    85,
      86,    87,    88,    89,    90,    91,    92,    93,    94,    95,
      96,    97,    98,    99,   100,   101,   102,   103,   104,   105,
     106,   107,   108,   109,   110,   111,   112,   113,   114,   115,
     116,   117,   118,   119,   120,   121,   122,   123,   124,   125,
     126,   127,   128,   129,   130,   131,   132,   133,   134,   135,
     136,   137,   138,   139,   140,   141,   143,   148,   149,   150,
     151,   152
};

#if YYDEBUG
  /* YYRLINE[YYN] -- Source line where rule number YYN was defined.  */
static const yytype_uint16 yyrline[] =
{
       0,   402,   402,   408,   414,   419,   424,   432,   433,   434,
     435,   436,   437,   438,   439,   440,   441,   442,   443,   444,
     445,   446,   447,   448,   449,   450,   451,   452,   453,   454,
     455,   458,   465,   472,   476,   480,   483,   486,   491,   495,
     499,   503,   508,   513,   520,   526,   530,   535,   541,   545,
     551,   558,   562,   567,   573,   576,   579,   582,   585,   588,
     591,   594,   599,   602,   605,   608,   611,   615,   619,   622,
     626,   630,   634,   637,   641,   645,   649,   652,   656,   659,
     663,   667,   671,   676,   681,   687,   693,   698,   713,   719,
     725,   732,   741,   752,   753,   756,   757,   758,   759,   760,
     761,   762,   763,   764,   765,   768,   769,   772,   781,   787,
     795,   803,   809,   817,   822,   830,   836,   844,   852,   858,
     867,   875,   882,   889,   897,   905,   912,   920,   936,   944,
     953,   958,   966,   970,   971,   974,   981,   987,   992,  1000,
    1004,  1005,  1012,  1016,  1017,  1025,  1031,  1039,  1049,  1050,
    1051,  1052,  1053,  1062,  1067,  1073,  1079,  1085,  1091,  1092,
    1093,  1095,  1097,  1098,  1099,  1101,  1105,  1110,  1121,  1126,
    1131,  1136,  1142,  1148,  1154,  1159,  1164,  1170,  1176,  1182,
    1188,  1196,  1202,  1208,  1226,  1232,  1238,  1243,  1249,  1255,
    1259,  1264,  1273,  1279,  1285,  1292,  1299,  1305,  1311,  1317,
    1323,  1329,  1335,  1341,  1347,  1353,  1359,  1365,  1371,  1377,
    1384,  1390,  1396,  1402,  1408,  1414,  1423,  1428,  1439,  1448,
    1456,  1463,  1470,  1477,  1484,  1491,  1498,  1505,  1512,  1517,
    1525,  1531,  1538,  1549,  1556,  1565,  1580,  1586,  1598,  1604,
    1613,  1618,  1624,  1632,  1640,  1646,  1653,  1660,  1666,  1672,
    1678,  1684,  1690,  1696,  1702,  1710,  1719,  1727,  1737,  1746,
    1757,  1766,  1774,  1780,  1788,  1796,  1806,  1812,  1817,  1825,
    1831,  1840,  1848,  1854,  1859,  1867,  1873,  1881,  1889,  1899,
    1905,  1913,  1920,  1928,  1935,  1949,  1961,  1971,  1972,  1978,
    1983
};
#endif

#if YYDEBUG || YYERROR_VERBOSE || 0
/* YYTNAME[SYMBOL-NUM] -- String name of the symbol SYMBOL-NUM.
   First, the terminals, then, starting at YYNTOKENS, nonterminals.  */
static const char *const yytname[] =
{
  "$end", "error", "$undefined", "\"DISTINCT\"", "\"AND\"", "\"ARRAY\"",
  "\"BOOLEAN\"", "\"DATATYPE\"", "\"ELSE\"", "\"ELSIF\"", "\"END\"",
  "\"ENDIF\"", "\"EXISTS\"", "\"FALSE\"", "\"FORALL\"", "\"IN\"", "\"IF\"",
  "\"LAMBDA\"", "\"SIMULATE\"", "\"LET\"", "\"NOT\"", "\"IS_INTEGER\"",
  "\"OF\"", "\"OR\"", "\"REAL\"", "\"INT\"", "\"SUBTYPE\"",
  "\"BITVECTOR\"", "\"THEN\"", "\"TRUE\"", "\"TYPE\"", "\"WITH\"",
  "\"XOR\"", "\"TCC\"", "\"PATTERN\"", "'_'", "DONE_TOK", "\"..\"",
  "\":=\"", "\"/=\"", "\"=>\"", "\"<=>\"", "\"+\"", "\"-\"", "\"*\"",
  "\"^\"", "\"/\"", "\"MOD\"", "\"DIV\"", "\"<\"", "\"<=\"", "\">\"",
  "\">=\"", "\"FLOOR\"", "\"<<\"", "\">>\"", "\"BVPLUS\"", "\"BVSUB\"",
  "\"BVUDIV\"", "\"BVSDIV\"", "\"BVUREM\"", "\"BVSREM\"", "\"BVSMOD\"",
  "\"BVSHL\"", "\"BVASHR\"", "\"BVLSHR\"", "\"BVUMINUS\"", "\"BVMULT\"",
  "\"[#\"", "\"#]\"", "\"(#\"", "\"#)\"", "\"->\"", "\"~\"", "\"&\"",
  "\"|\"", "\"BVXOR\"", "\"BVNAND\"", "\"BVNOR\"", "\"BVCOMP\"",
  "\"BVXNOR\"", "\"@\"", "\"BVTOINT\"", "\"INTTOBV\"", "\"BOOLEXTRACT\"",
  "\"BVLT\"", "\"BVGT\"", "\"BVLE\"", "\"BVGE\"", "\"SX\"",
  "\"BVZEROEXTEND\"", "\"BVREPEAT\"", "\"BVROTL\"", "\"BVROTR\"",
  "\"BVSLT\"", "\"BVSGT\"", "\"BVSLE\"", "\"BVSGE\"",
  "\"ARITH_VAR_ORDER\"", "\"ASSERT\"", "\"QUERY\"", "\"CHECKSAT\"",
  "\"CONTINUE\"", "\"RESTART\"", "\"HELP\"", "\"DBG\"", "\"TRACE\"",
  "\"UNTRACE\"", "\"OPTION\"", "\"TRANSFORM\"", "\"PRINT\"",
  "\"PRINT_TYPE\"", "\"CALL\"", "\"ECHO\"", "\"INCLUDE\"",
  "\"DUMP_PROOF\"", "\"DUMP_ASSUMPTIONS\"", "\"DUMP_SIG\"", "\"DUMP_TCC\"",
  "\"DUMP_TCC_ASSUMPTIONS\"", "\"DUMP_TCC_PROOF\"", "\"DUMP_CLOSURE\"",
  "\"DUMP_CLOSURE_PROOF\"", "\"WHERE\"", "\"ASSERTIONS\"",
  "\"ASSUMPTIONS\"", "\"COUNTEREXAMPLE\"", "\"COUNTERMODEL\"", "\"PUSH\"",
  "\"POP\"", "\"POPTO\"", "\"PUSH_SCOPE\"", "\"POP_SCOPE\"",
  "\"POPTO_SCOPE\"", "\"RESET\"", "\"CONTEXT\"", "\"FORGET\"",
  "\"GET_TYPE\"", "\"CHECK_TYPE\"", "\"GET_CHILD\"", "\"GET_OP\"",
  "\"SUBSTITUTE\"", "'='", "UMINUS_TOK", "'['", "'{'", "'.'", "'('",
  "ID_TOK", "STRINGLIT_TOK", "NUMERAL_TOK", "HEX_TOK", "BINARY_TOK", "';'",
  "')'", "':'", "']'", "','", "'}'", "$accept", "cmd", "other_cmd",
  "Arith_Var_Order", "Assert", "Query", "Dbg", "Trace", "Option", "Help",
  "Transform", "Print", "Call", "Echo", "Include", "DumpCommand", "Where",
  "Push", "Pop", "PopTo", "Context", "Forget", "Get_Type", "Check_Type",
  "Get_Child", "Get_Op", "Substitute", "Identifier", "StringLiteral",
  "Numeral", "Binary", "Hex", "Type", "TypeNotIdentifier", "TypeDef",
  "DataType", "SingleDataTypes", "SingleDataType", "Constructors",
  "Constructor", "VarDecls", "VarDecl", "FunctionType", "RecordType",
  "FieldDecls", "FieldDecl", "TupleType", "TypeList", "ArrayType",
  "ScalarType", "SubType", "BasicType", "BitvectorType", "Real", "Int",
  "SubrangeType", "LeftBound", "RightBound", "reverseIDs", "IDs", "Expr",
  "AndExpr", "OrExpr", "Conditional", "ElseRest", "Exprs", "Pattern",
  "Patterns", "Update", "Updates", "UpdatePositionNode", "UpdatePosition",
  "Lambda", "QuantExpr", "ArrayLiteral", "RecordLiteral", "RecordEntries",
  "RecordEntry", "TupleLiteral", "LetDeclsNode", "LetDecls", "LetDecl",
  "LetExpr", "TypeLetDeclsNode", "TypeLetDecls", "TypeLetDecl",
  "TypeLetExpr", "BoundVarDecl", "BoundVarDecls", "BoundVarDeclNode",
  "ConstDecl", "TypeDecl", YY_NULLPTR
};
#endif

# ifdef YYPRINT
/* YYTOKNUM[NUM] -- (External) token number corresponding to the
   (internal) symbol number NUM (which must be that of a token).  */
static const yytype_uint16 yytoknum[] =
{
       0,   256,   257,   258,   259,   260,   261,   262,   263,   264,
     265,   266,   267,   268,   269,   270,   271,   272,   273,   274,
     275,   276,   277,   278,   279,   280,   281,   282,   283,   284,
     285,   286,   287,   288,   289,    95,   290,   291,   292,   293,
     294,   295,   296,   297,   298,   299,   300,   301,   302,   303,
     304,   305,   306,   307,   308,   309,   310,   311,   312,   313,
     314,   315,   316,   317,   318,   319,   320,   321,   322,   323,
     324,   325,   326,   327,   328,   329,   330,   331,   332,   333,
     334,   335,   336,   337,   338,   339,   340,   341,   342,   343,
     344,   345,   346,   347,   348,   349,   350,   351,   352,   353,
     354,   355,   356,   357,   358,   359,   360,   361,   362,   363,
     364,   365,   366,   367,   368,   369,   370,   371,   372,   373,
     374,   375,   376,   377,   378,   379,   380,   381,   382,   383,
     384,   385,   386,   387,   388,   389,   390,   391,   392,   393,
     394,   395,    61,   396,    91,   123,    46,    40,   397,   398,
     399,   400,   401,    59,    41,    58,    93,    44,   125
};
# endif

#define YYPACT_NINF -512

#define yypact_value_is_default(Yystate) \
  (!!((Yystate) == (-512)))

#define YYTABLE_NINF -1

#define yytable_value_is_error(Yytable_value) \
  (!!((Yytable_value) == (-1)))

  /* YYPACT[STATE-NUM] -- Index in YYTABLE of the portion describing
     STATE-NUM.  */
static const yytype_int16 yypact[] =
{
    4889,  -121,  -512,  -128,  4688,  4688,  4688,  -512,  4688,  -112,
    -112,  -112,  -112,  -112,  4688,  4688,   384,  -121,  -112,  -112,
    -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,
    -512,  -512,  -512,   -91,   -91,   -91,   -91,   -91,   -91,  -512,
    -112,  -121,  4688,  4688,  4688,  4688,  -121,  -512,  -512,    76,
     -68,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,
    -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,
    -512,  -512,  -512,  -512,  -512,  -129,  -512,   -63,   -42,   -22,
      -5,  -512,  4688,   -20,     1,     7,  -512,    11,  4688,    30,
      31,  -121,  4688,  4688,  -512,    33,  4688,    34,    36,    37,
      38,    39,    40,    42,    43,    51,    52,    54,    55,    58,
    -121,  4688,    59,    61,    65,    71,    74,    75,    77,    79,
      80,    91,    93,    94,   116,   117,   120,   134,   135,   141,
     142,   144,   146,  4688,  -512,  -512,  -512,  -512,  -512,  -512,
    -512,  5306,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,
    -512,  5306,  5306,  5306,  -512,  -512,  -512,  -512,  -512,   -36,
    5306,  5306,   384,  -512,  -121,  -512,  -512,   149,   150,  -121,
      44,   384,  -512,    78,  -512,  -512,  -512,  -512,  -512,  -512,
    -512,  -512,  -512,  -512,  -512,  -512,  4688,  -512,  -512,  -512,
    -512,  -512,  -512,  -512,  -512,  -512,  -512,  5306,  2680,  5027,
    5306,   145,  -512,  -512,  -121,    56,  -121,  -512,  -512,  -121,
    -512,  -121,  5306,  -106,  4688,  -121,  -121,  -121,  5144,  -121,
    4688,   -98,   284,   147,  -512,  5521,     9,  4688,   295,  4688,
     153,   155,  4688,  4688,  4688,  4688,  4688,  4688,  4688,  4688,
    4688,   157,   272,   -62,  -512,   -25,  4688,  4688,  4688,  4688,
    4688,  4688,  4688,  4688,  4688,  4688,  4688,  4688,  4688,  4688,
    4688,  4688,  4688,  4688,  4688,  4688,  4688,  2733,   152,  4688,
    4688,   -14,  4688,  4688,  4688,  4688,  4688,  4688,  4788,  4688,
    4688,  4688,  4688,  4688,  4688,   161,   162,  4688,  4688,  4688,
    4688,  4888,    15,  4688,   -91,  -512,  -512,     0,   -95,   306,
     165,  -512,  4688,   -91,   168,   -52,  -512,  -512,   -91,  -512,
     -47,   -92,   290,   -29,  -102,   384,  5306,   384,  -512,   384,
     175,  -512,   178,  -512,   177,   183,   199,   -49,   188,   204,
     278,  -512,  -512,  -512,  4688,   -48,   200,   -21,   -15,  4688,
      10,    12,  4688,   384,  4688,  -121,  2797,  2850,   198,   201,
     816,   869,   925,   989,  1045,  1098,  1154,  1207,  2914,   202,
    4688,  -512,  -121,  1263,  1316,  1372,  1425,  1481,  2967,  1534,
    1590,  1643,  1699,  1752,  1808,  1861,  1917,  1970,  2026,  2079,
    2135,  2188,  2244,  2297,  -512,  4688,  5521,  5468,  4688,    15,
    -512,  -512,   203,   323,     3,  5468,  5574,  5415,  5415,   375,
     375,  4688,    63,   271,    63,   603,   603,   603,   603,  -512,
    -512,   -25,    49,    -1,  5574,   207,  2462,  -512,  -512,    16,
    -512,   384,   384,   335,   384,  -121,   551,   212,   384,  -512,
    -121,  -512,   384,   384,  -512,   384,   -19,  -512,   296,  -512,
      78,   -27,  -121,   384,  -121,   216,    72,  4688,   289,  -121,
    -121,  5306,  -512,   217,   218,   219,  5080,   220,  -512,  5306,
     -26,  5306,  -512,  -512,  -512,  4688,  4688,  4688,  4688,  4688,
    4688,  4688,  4688,  4688,  4688,  -512,  4688,  5306,  -512,  4688,
    4688,  4688,  4688,  4688,  -512,   227,   228,  4688,  4688,  4688,
    4688,   229,   230,   232,   233,   235,  4688,  4688,  4688,  4688,
    3031,  2515,  -512,  -512,   -14,  4688,  4688,    15,  -512,   671,
     241,  -512,  -512,    78,    78,   225,    78,  -512,  -512,  4688,
    -512,    78,  -512,   236,    78,    78,  -512,   -91,  -512,   238,
     384,  4688,  -512,    78,  -512,   384,  -121,    78,  -512,  -512,
    5306,  -512,    78,   240,    17,  -512,  -512,  4688,  4588,  4588,
    4688,  4688,  -512,  4688,  4688,    18,  2353,  3084,  3148,  3201,
    3265,  3318,  3382,  3435,  3499,  2406,  3552,  3616,  3669,  3733,
    3786,   239,   244,  3850,  3903,  3967,  4020,   245,   246,   247,
     248,   250,  4084,  4137,  4201,  4254,  -512,  -512,  -512,  5306,
    2571,  -512,  -512,  4688,   249,   384,  4318,  -512,  -512,  -512,
    -512,  5362,   -16,   254,   384,  -512,  -121,  5306,   260,  5306,
    -512,  4588,  5306,  4588,  5197,  5253,  5306,  5306,  -512,  4688,
    -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  4688,  -512,
    -512,  -512,  -512,  -512,   263,  -512,  -512,  -512,  -512,  -512,
    -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,
    4371,  -512,    78,  -512,  4888,  4688,  -512,    78,  -512,  4688,
    5306,  -512,  5306,  -512,  4688,  4435,  4488,   262,  -512,   376,
    5306,    22,  5080,  -512,  -512,  -512,  4688,   267,  -512,  2624,
    -512,  -512
};

  /* YYDEFACT[STATE-NUM] -- Default reduction number in state STATE-NUM.
     Performed when YYTABLE does not specify something else to do.  Zero
     means the default is an error.  */
static const yytype_uint16 yydefact[] =
{
       0,     0,     6,     0,     0,     0,    35,    36,     0,    46,
      39,     0,     0,     0,     0,     0,     0,     0,    52,     0,
      54,    55,    56,    57,    58,    59,    60,    61,    62,    63,
      64,    65,    66,    68,    72,    76,    70,    74,    78,    79,
      81,     0,     0,     0,     0,     0,     0,    88,     5,     0,
       0,    30,     7,     8,     9,    10,    11,    12,    13,    14,
      15,    16,    17,    18,    19,    20,    21,    22,    23,    29,
      27,    28,    25,    26,    24,     0,   287,     0,     0,     0,
       0,   108,     0,     0,     0,     0,   166,     0,     0,     0,
       0,     0,     0,     0,   165,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,    90,    92,    91,   148,   149,   150,
     151,    32,   175,   174,   164,   158,   159,   161,   162,   163,
     160,    33,    34,    37,    89,    45,    38,    40,    41,     0,
      47,    48,     0,   132,     0,   136,   137,     0,     0,     0,
       0,     0,    93,    49,    94,    96,   100,    99,    95,   103,
      97,   102,   133,   134,    98,   101,     0,    51,    53,    67,
      71,    75,    69,    73,    77,    80,    82,    83,     0,     0,
      86,     0,     1,     4,     0,     0,     0,     3,     2,     0,
     107,     0,   238,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,   266,   267,   168,   169,     0,   167,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,   262,   190,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,   238,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,    44,    42,     0,     0,     0,
     272,   273,     0,     0,     0,     0,   122,   139,     0,   140,
       0,     0,     0,     0,     0,     0,    50,     0,    85,     0,
     145,   147,     0,   279,   281,     0,   289,   283,     0,   113,
     110,   111,   109,    31,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   261,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,   189,     0,   232,   234,     0,     0,
     248,   244,   157,     0,   246,   173,   172,   176,   177,   178,
     179,     0,   181,   182,   183,   187,   188,   184,   185,   194,
     195,   192,   193,   214,   171,    90,     0,   155,   156,     0,
      43,     0,     0,     0,     0,     0,     0,     0,     0,   121,
       0,   141,     0,     0,   125,     0,     0,   104,     0,   119,
      84,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   239,   230,     0,     0,     0,     0,     0,   153,   269,
       0,   271,   268,   170,   186,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,   228,     0,   264,   263,     0,
       0,     0,     0,     0,   216,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     239,     0,   249,   250,     0,     0,     0,     0,   252,     0,
       0,   154,   152,   128,   275,     0,   277,   274,   130,     0,
     135,   124,   123,   119,   126,   127,   142,     0,   143,     0,
       0,     0,   146,   278,   280,     0,     0,   105,   288,   106,
     282,   290,   286,     0,     0,   115,   112,     0,     0,     0,
       0,     0,   235,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,   265,   247,   245,   243,
       0,   253,   254,     0,     0,     0,     0,   118,   144,   138,
     120,     0,   285,     0,     0,   114,     0,   260,     0,   258,
     241,     0,   256,     0,     0,     0,   255,   270,   218,     0,
     220,   221,   222,   223,   224,   225,   226,   227,     0,   196,
     197,   198,   199,   200,     0,   217,   206,   207,   208,   209,
     201,   202,   203,   204,   205,   210,   211,   212,   213,   251,
       0,   191,   276,   131,     0,     0,   129,   117,   116,     0,
     259,   242,   257,   236,     0,     0,     0,     0,   180,   148,
     284,     0,     0,   219,   229,   215,     0,     0,   237,     0,
     240,    87
};

  /* YYPGOTO[NTERM-NUM].  */
static const yytype_int16 yypgoto[] =
{
    -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,
    -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,
    -512,  -512,  -512,  -512,  -512,  -512,  -512,   347,     2,    -2,
    -512,  -512,  -104,  -512,  -512,  -512,  -512,   213,  -512,   -23,
    -512,  -183,  -512,  -512,  -512,     5,  -512,   257,  -512,  -512,
    -512,  -512,  -512,  -512,  -512,  -512,  -512,  -512,   -17,  -200,
      -4,  -512,  -512,  -512,  -240,   -79,  -511,  -109,   -60,  -512,
    -512,  -512,  -512,  -512,  -512,  -512,  -512,    83,  -512,  -512,
    -512,   101,  -512,  -512,  -512,    23,  -512,  -207,   -57,  -512,
    -512,  -512
};

  /* YYDEFGOTO[NTERM-NUM].  */
static const yytype_int16 yydefgoto[] =
{
      -1,    49,    50,    51,    52,    53,    54,    55,    56,    57,
      58,    59,    60,    61,    62,    63,    64,    65,    66,    67,
      68,    69,    70,    71,    72,    73,    74,   137,   155,   138,
     139,   140,   173,   174,   538,    76,    80,    81,   330,   331,
     544,   545,   175,   176,   305,   306,   177,   311,   178,   539,
     179,   180,   181,   182,   183,   184,   312,   529,   321,   322,
     212,   142,   143,   144,   552,   268,   610,   611,   391,   392,
     393,   394,   145,   146,   147,   148,   243,   244,   149,   222,
     223,   224,   150,   299,   300,   301,   185,   323,   324,   325,
      77,    78
};

  /* YYTABLE[YYPACT[STATE-NUM]] -- What to do in state STATE-NUM.  If
     positive, shift that token.  If negative, reduce the rule whose
     number is the opposite.  If YYTABLE_NINF, syntax error.  */
static const yytype_int16 yytable[] =
{
     141,   151,   152,   213,   153,   210,   328,   294,   336,   361,
     160,   161,   156,   157,   158,   159,   526,   429,   204,    82,
     187,   188,   421,   315,   527,   432,   205,    47,   206,   285,
     286,   189,   190,   191,   192,   193,   194,   154,   197,   198,
     199,   200,   195,   315,   342,   315,   315,   422,   333,   162,
     163,   334,   438,   285,   286,   435,   315,   343,   297,   134,
     423,   162,   163,   164,   434,   435,   310,   313,   165,   166,
     167,   168,   315,   287,   288,   164,   202,   162,   163,   307,
     165,   166,   167,   168,   218,   203,   326,   308,   225,   226,
     207,   164,   228,   447,   271,   362,   165,   166,   167,   168,
     661,   327,   661,   285,   286,   430,   452,   245,   279,   334,
     433,   208,   169,   154,   134,   531,   554,   285,   286,   291,
     209,   292,   293,   287,   169,   437,   655,   214,   433,   267,
     388,   134,   389,   454,    47,   335,   444,   287,   288,   455,
     169,   341,   444,   291,   289,   292,   293,   506,   215,   507,
     315,    47,   211,   291,   216,   292,   293,   296,   217,   337,
     338,   295,   340,    47,   457,   134,   458,   444,   309,   334,
     512,   605,   618,   334,   606,   334,   677,   219,   220,   334,
     227,   229,   316,   230,   231,   232,   233,   234,   170,   235,
     236,   171,    47,   291,   134,   292,   293,   318,   237,   238,
     170,   239,   240,   171,    47,   241,   246,   291,   247,   292,
     293,   439,   248,   440,   419,   441,   170,   536,   249,   171,
      47,   250,   251,   346,   252,   347,   253,   254,   350,   351,
     352,   353,   354,   355,   356,   357,   358,   534,   255,   460,
     256,   257,   363,   364,   365,   366,   367,   368,   369,   370,
     371,   372,   373,   374,   375,   376,   377,   378,   379,   380,
     381,   382,   383,   258,   259,   386,   387,   260,   395,   396,
     397,   398,   399,   400,   402,   403,   404,   405,   406,   407,
     408,   261,   262,   411,   412,   413,   414,   416,   263,   264,
     418,   265,   420,   266,   162,   163,   302,   303,   426,   344,
     319,   427,   271,   348,   345,   349,   431,   359,   164,   385,
     360,   409,   410,   165,   166,   167,   168,   513,   514,   541,
     516,   424,   425,   428,   521,   285,   286,   436,   523,   524,
     451,   525,   442,   443,   444,   456,   603,   445,   459,   533,
     461,   446,   537,   448,   542,   287,   288,    75,    79,   285,
     286,   449,   289,   450,   453,   465,   477,   169,   466,   476,
     504,   505,   510,   172,   186,   515,   520,   595,   530,   287,
     288,   535,   547,   548,   549,   553,   289,   571,   572,   577,
     578,   500,   579,   580,   501,   581,   555,   503,   196,   162,
     163,   594,   597,   201,   599,   604,   634,   509,   635,   640,
     641,   642,   643,   164,   644,   651,   271,   659,   165,   166,
     167,   168,   656,   667,   676,   291,   675,   292,   293,   278,
     279,   280,   680,   658,   332,   532,   600,   546,   314,   285,
     286,   602,   678,   170,   528,   522,   171,    47,   221,   291,
     613,   292,   293,   540,   588,   478,   462,     0,   517,   287,
     288,     0,   169,     0,     0,     0,   289,   242,     0,     0,
       0,     0,   556,   557,   558,   559,   560,   561,   562,   563,
     564,     0,   565,     0,     0,   566,   567,   568,   569,   570,
       0,     0,     0,   573,   574,   575,   576,     0,     0,     0,
       0,   652,   582,   583,   584,   585,     0,     0,     0,     0,
     657,   589,   590,     0,     0,   592,     0,     0,     0,   172,
       0,   298,     0,     0,     0,   596,   304,   172,   172,   291,
       0,   292,   293,     0,     0,   598,     0,   601,   170,     0,
       0,   171,    47,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,   607,   609,   612,   614,   615,     0,   616,
     617,   320,   172,   320,     0,   269,   329,     0,    79,     0,
       0,     0,   320,   320,   320,     0,   320,     0,     0,     0,
       0,     0,     0,     0,   270,     0,     0,     0,     0,     0,
     671,     0,   271,   272,     0,     0,     0,     0,     0,   650,
     273,   274,   275,   276,   277,   278,   279,   280,     0,     0,
     281,   282,   283,   284,     0,   285,   286,   660,     0,   662,
       0,     0,     0,     0,     0,   665,     0,     0,   390,     0,
       0,     0,     0,     0,   666,   287,   288,     0,     0,     0,
       0,     0,   289,     0,   271,     0,     0,     0,     0,   417,
       0,     0,     0,     0,     0,   276,   277,   278,   279,   280,
     416,   670,    -1,    -1,    -1,    -1,     0,   285,   286,     0,
     672,     0,   172,     0,   172,     0,   172,     0,     0,     0,
       0,     0,   679,     0,     0,   269,     0,   287,   288,     0,
       0,     0,     0,     0,   289,     0,     0,     0,     0,     0,
     172,     0,   221,   290,   270,   291,     0,   292,   293,     0,
       0,     0,   271,   272,     0,   518,     0,     0,   519,   242,
     273,   274,   275,   276,   277,   278,   279,   280,     0,     0,
     281,   282,   283,   284,     0,   285,   286,     0,     0,     0,
       0,     0,     0,     0,     0,     0,   502,     0,     0,     0,
       0,   508,     0,     0,     0,   287,   288,   291,     0,   292,
     293,     0,   289,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   172,   172,
       0,   172,   298,     0,     0,   172,     0,   304,     0,   172,
     172,     0,   172,     0,     0,     0,     0,     0,     0,   320,
     172,   320,     0,   172,     0,   172,   543,   329,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,   290,     0,   291,     0,   292,   293,     0,
     269,     0,     0,     0,     0,   384,     0,     0,   593,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   270,
       0,     0,     0,     0,     0,     0,     0,   271,   272,     0,
       0,   390,     0,     0,   591,   273,   274,   275,   276,   277,
     278,   279,   280,     0,     0,   281,   282,   283,   284,     0,
     285,   286,     0,   269,     0,     0,     0,   172,     0,     0,
       0,     0,   172,   320,     0,     0,     0,     0,     0,     0,
     287,   288,   270,     0,     0,     0,     0,   289,     0,     0,
     271,   272,     0,     0,     0,     0,     0,     0,   273,   274,
     275,   276,   277,   278,   279,   280,     0,     0,   281,   282,
     283,   284,     0,   285,   286,     0,     0,     0,     0,   269,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   172,   287,   288,     0,     0,     0,   270,     0,
     289,   172,     0,   543,     0,     0,   271,   272,   290,     0,
     291,     0,   292,   293,   273,   274,   275,   276,   277,   278,
     279,   280,     0,   467,   281,   282,   283,   284,     0,   285,
     286,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,   269,     0,     0,     0,     0,     0,   287,
     288,   669,     0,     0,     0,     0,   289,     0,     0,     0,
       0,   290,   270,   291,     0,   292,   293,     0,     0,     0,
     271,   272,     0,     0,     0,     0,   468,     0,   273,   274,
     275,   276,   277,   278,   279,   280,     0,     0,   281,   282,
     283,   284,     0,   285,   286,     0,     0,     0,     0,   269,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,   287,   288,     0,     0,   290,   270,   291,
     289,   292,   293,     0,     0,     0,   271,   272,     0,     0,
       0,     0,   469,     0,   273,   274,   275,   276,   277,   278,
     279,   280,     0,     0,   281,   282,   283,   284,     0,   285,
     286,     0,   269,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   287,
     288,   270,     0,     0,     0,     0,   289,     0,     0,   271,
     272,   290,     0,   291,     0,   292,   293,   273,   274,   275,
     276,   277,   278,   279,   280,     0,   470,   281,   282,   283,
     284,     0,   285,   286,     0,     0,     0,     0,   269,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   287,   288,     0,     0,     0,   270,     0,   289,
       0,     0,     0,     0,     0,   271,   272,   290,     0,   291,
       0,   292,   293,   273,   274,   275,   276,   277,   278,   279,
     280,     0,   471,   281,   282,   283,   284,     0,   285,   286,
       0,   269,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   287,   288,
     270,     0,     0,     0,     0,   289,     0,     0,   271,   272,
     290,     0,   291,     0,   292,   293,   273,   274,   275,   276,
     277,   278,   279,   280,     0,   472,   281,   282,   283,   284,
       0,   285,   286,     0,     0,     0,     0,   269,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   287,   288,     0,     0,     0,   270,     0,   289,     0,
       0,     0,     0,     0,   271,   272,   290,     0,   291,     0,
     292,   293,   273,   274,   275,   276,   277,   278,   279,   280,
       0,   473,   281,   282,   283,   284,     0,   285,   286,     0,
     269,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,   287,   288,   270,
       0,     0,     0,     0,   289,     0,     0,   271,   272,   290,
       0,   291,     0,   292,   293,   273,   274,   275,   276,   277,
     278,   279,   280,     0,   474,   281,   282,   283,   284,     0,
     285,   286,     0,     0,     0,     0,   269,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     287,   288,     0,     0,     0,   270,     0,   289,     0,     0,
       0,     0,     0,   271,   272,   290,     0,   291,     0,   292,
     293,   273,   274,   275,   276,   277,   278,   279,   280,     0,
     479,   281,   282,   283,   284,     0,   285,   286,     0,   269,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,   287,   288,   270,     0,
       0,     0,     0,   289,     0,     0,   271,   272,   290,     0,
     291,     0,   292,   293,   273,   274,   275,   276,   277,   278,
     279,   280,     0,   480,   281,   282,   283,   284,     0,   285,
     286,     0,     0,     0,     0,   269,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   287,
     288,     0,     0,     0,   270,     0,   289,     0,     0,     0,
       0,     0,   271,   272,   290,     0,   291,     0,   292,   293,
     273,   274,   275,   276,   277,   278,   279,   280,     0,   481,
     281,   282,   283,   284,     0,   285,   286,     0,   269,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,   287,   288,   270,     0,     0,
       0,     0,   289,     0,     0,   271,   272,   290,     0,   291,
       0,   292,   293,   273,   274,   275,   276,   277,   278,   279,
     280,     0,   482,   281,   282,   283,   284,     0,   285,   286,
       0,     0,     0,     0,   269,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   287,   288,
       0,     0,     0,   270,     0,   289,     0,     0,     0,     0,
       0,   271,   272,   290,     0,   291,     0,   292,   293,   273,
     274,   275,   276,   277,   278,   279,   280,     0,   483,   281,
     282,   283,   284,     0,   285,   286,     0,   269,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,   287,   288,   270,     0,     0,     0,
       0,   289,     0,     0,   271,   272,   290,     0,   291,     0,
     292,   293,   273,   274,   275,   276,   277,   278,   279,   280,
       0,   485,   281,   282,   283,   284,     0,   285,   286,     0,
       0,     0,     0,   269,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,   287,   288,     0,
       0,     0,   270,     0,   289,     0,     0,     0,     0,     0,
     271,   272,   290,     0,   291,     0,   292,   293,   273,   274,
     275,   276,   277,   278,   279,   280,     0,   486,   281,   282,
     283,   284,     0,   285,   286,     0,   269,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,   287,   288,   270,     0,     0,     0,     0,
     289,     0,     0,   271,   272,   290,     0,   291,     0,   292,
     293,   273,   274,   275,   276,   277,   278,   279,   280,     0,
     487,   281,   282,   283,   284,     0,   285,   286,     0,     0,
       0,     0,   269,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,   287,   288,     0,     0,
       0,   270,     0,   289,     0,     0,     0,     0,     0,   271,
     272,   290,     0,   291,     0,   292,   293,   273,   274,   275,
     276,   277,   278,   279,   280,     0,   488,   281,   282,   283,
     284,     0,   285,   286,     0,   269,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   287,   288,   270,     0,     0,     0,     0,   289,
       0,     0,   271,   272,   290,     0,   291,     0,   292,   293,
     273,   274,   275,   276,   277,   278,   279,   280,     0,   489,
     281,   282,   283,   284,     0,   285,   286,     0,     0,     0,
       0,   269,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,   287,   288,     0,     0,     0,
     270,     0,   289,     0,     0,     0,     0,     0,   271,   272,
     290,     0,   291,     0,   292,   293,   273,   274,   275,   276,
     277,   278,   279,   280,     0,   490,   281,   282,   283,   284,
       0,   285,   286,     0,   269,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   287,   288,   270,     0,     0,     0,     0,   289,     0,
       0,   271,   272,   290,     0,   291,     0,   292,   293,   273,
     274,   275,   276,   277,   278,   279,   280,     0,   491,   281,
     282,   283,   284,     0,   285,   286,     0,     0,     0,     0,
     269,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,   287,   288,     0,     0,     0,   270,
       0,   289,     0,     0,     0,     0,     0,   271,   272,   290,
       0,   291,     0,   292,   293,   273,   274,   275,   276,   277,
     278,   279,   280,     0,   492,   281,   282,   283,   284,     0,
     285,   286,     0,   269,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     287,   288,   270,     0,     0,     0,     0,   289,     0,     0,
     271,   272,   290,     0,   291,     0,   292,   293,   273,   274,
     275,   276,   277,   278,   279,   280,     0,   493,   281,   282,
     283,   284,     0,   285,   286,     0,     0,     0,     0,   269,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,   287,   288,     0,     0,     0,   270,     0,
     289,     0,     0,     0,     0,     0,   271,   272,   290,     0,
     291,     0,   292,   293,   273,   274,   275,   276,   277,   278,
     279,   280,     0,   494,   281,   282,   283,   284,     0,   285,
     286,     0,   269,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   287,
     288,   270,     0,     0,     0,     0,   289,     0,     0,   271,
     272,   290,     0,   291,     0,   292,   293,   273,   274,   275,
     276,   277,   278,   279,   280,     0,   495,   281,   282,   283,
     284,     0,   285,   286,     0,     0,     0,     0,   269,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   287,   288,     0,     0,     0,   270,     0,   289,
       0,     0,     0,     0,     0,   271,   272,   290,     0,   291,
       0,   292,   293,   273,   274,   275,   276,   277,   278,   279,
     280,     0,   496,   281,   282,   283,   284,     0,   285,   286,
       0,   269,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   287,   288,
     270,     0,     0,     0,     0,   289,     0,     0,   271,   272,
     290,     0,   291,     0,   292,   293,   273,   274,   275,   276,
     277,   278,   279,   280,     0,   497,   281,   282,   283,   284,
       0,   285,   286,     0,     0,     0,     0,   269,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   287,   288,     0,     0,     0,   270,     0,   289,     0,
       0,     0,     0,     0,   271,   272,   290,     0,   291,     0,
     292,   293,   273,   274,   275,   276,   277,   278,   279,   280,
       0,   498,   281,   282,   283,   284,     0,   285,   286,     0,
     269,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,   287,   288,   270,
       0,     0,     0,     0,   289,     0,     0,   271,   272,   290,
       0,   291,     0,   292,   293,   273,   274,   275,   276,   277,
     278,   279,   280,     0,   499,   281,   282,   283,   284,     0,
     285,   286,     0,     0,     0,     0,   269,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     287,   288,     0,     0,     0,   270,     0,   289,     0,     0,
       0,     0,     0,   271,   272,   290,     0,   291,     0,   292,
     293,   273,   274,   275,   276,   277,   278,   279,   280,     0,
     619,   281,   282,   283,   284,     0,   285,   286,     0,   269,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,   287,   288,   270,     0,
       0,     0,     0,   289,     0,     0,   271,   272,   290,     0,
     291,     0,   292,   293,   273,   274,   275,   276,   277,   278,
     279,   280,     0,   628,   281,   282,   283,   284,     0,   285,
     286,     0,     0,     0,     0,   269,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   287,
     288,     0,     0,     0,   270,     0,   289,     0,     0,     0,
       0,     0,   271,   272,   290,     0,   291,     0,   292,   293,
     273,   274,   275,   276,   277,   278,   279,   280,   511,     0,
     281,   282,   283,   284,     0,   285,   286,     0,   269,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,   287,   288,   270,     0,     0,
       0,     0,   289,     0,     0,   271,   272,   290,     0,   291,
       0,   292,   293,   273,   274,   275,   276,   277,   278,   279,
     280,   587,     0,   281,   282,   283,   284,     0,   285,   286,
       0,     0,     0,     0,   269,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   287,   288,
       0,     0,     0,   270,     0,   289,     0,     0,     0,     0,
       0,   271,   272,   290,     0,   291,     0,   292,   293,   273,
     274,   275,   276,   277,   278,   279,   280,   649,     0,   281,
     282,   283,   284,     0,   285,   286,     0,   269,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,   287,   288,   270,     0,     0,     0,
       0,   289,     0,     0,   271,   272,   290,     0,   291,     0,
     292,   293,   273,   274,   275,   276,   277,   278,   279,   280,
     681,     0,   281,   282,   283,   284,     0,   285,   286,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   269,     0,     0,     0,     0,     0,   287,   288,     0,
       0,     0,     0,     0,   289,     0,     0,     0,     0,     0,
     270,     0,   290,     0,   291,     0,   292,   293,   271,   272,
       0,     0,     0,     0,     0,   317,   273,   274,   275,   276,
     277,   278,   279,   280,     0,     0,   281,   282,   283,   284,
       0,   285,   286,     0,   269,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   287,   288,   270,     0,   290,     0,   291,   289,   292,
     293,   271,   272,     0,     0,     0,     0,   384,     0,   273,
     274,   275,   276,   277,   278,   279,   280,     0,     0,   281,
     282,   283,   284,     0,   285,   286,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   269,     0,
       0,     0,     0,     0,   287,   288,     0,     0,     0,     0,
       0,   289,     0,     0,     0,     0,     0,   270,     0,   290,
       0,   291,     0,   292,   293,   271,   272,     0,     0,     0,
       0,   463,     0,   273,   274,   275,   276,   277,   278,   279,
     280,     0,     0,   281,   282,   283,   284,     0,   285,   286,
       0,   269,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   287,   288,
     270,     0,   290,     0,   291,   289,   292,   293,   271,   272,
       0,     0,     0,     0,   464,     0,   273,   274,   275,   276,
     277,   278,   279,   280,     0,     0,   281,   282,   283,   284,
       0,   285,   286,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,   269,     0,     0,     0,     0,
       0,   287,   288,     0,     0,     0,     0,     0,   289,     0,
       0,     0,     0,     0,   270,     0,   290,     0,   291,     0,
     292,   293,   271,   272,     0,     0,     0,     0,   475,     0,
     273,   274,   275,   276,   277,   278,   279,   280,     0,     0,
     281,   282,   283,   284,     0,   285,   286,     0,   269,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,   287,   288,   270,     0,   290,
       0,   291,   289,   292,   293,   271,   272,     0,     0,     0,
       0,   484,     0,   273,   274,   275,   276,   277,   278,   279,
     280,     0,     0,   281,   282,   283,   284,     0,   285,   286,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   269,     0,     0,     0,     0,     0,   287,   288,
       0,     0,     0,     0,     0,   289,     0,     0,     0,     0,
       0,   270,     0,   290,     0,   291,     0,   292,   293,   271,
     272,     0,     0,     0,     0,   586,     0,   273,   274,   275,
     276,   277,   278,   279,   280,     0,     0,   281,   282,   283,
     284,     0,   285,   286,     0,   269,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   287,   288,   270,     0,   290,     0,   291,   289,
     292,   293,   271,   272,     0,     0,     0,     0,   620,     0,
     273,   274,   275,   276,   277,   278,   279,   280,     0,     0,
     281,   282,   283,   284,     0,   285,   286,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   269,
       0,     0,     0,     0,     0,   287,   288,     0,     0,     0,
       0,     0,   289,     0,     0,     0,     0,     0,   270,     0,
     290,     0,   291,     0,   292,   293,   271,   272,     0,     0,
       0,     0,   621,     0,   273,   274,   275,   276,   277,   278,
     279,   280,     0,     0,   281,   282,   283,   284,     0,   285,
     286,     0,   269,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   287,
     288,   270,     0,   290,     0,   291,   289,   292,   293,   271,
     272,     0,     0,     0,     0,   622,     0,   273,   274,   275,
     276,   277,   278,   279,   280,     0,     0,   281,   282,   283,
     284,     0,   285,   286,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,   269,     0,     0,     0,
       0,     0,   287,   288,     0,     0,     0,     0,     0,   289,
       0,     0,     0,     0,     0,   270,     0,   290,     0,   291,
       0,   292,   293,   271,   272,     0,     0,     0,     0,   623,
       0,   273,   274,   275,   276,   277,   278,   279,   280,     0,
       0,   281,   282,   283,   284,     0,   285,   286,     0,   269,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,   287,   288,   270,     0,
     290,     0,   291,   289,   292,   293,   271,   272,     0,     0,
       0,     0,   624,     0,   273,   274,   275,   276,   277,   278,
     279,   280,     0,     0,   281,   282,   283,   284,     0,   285,
     286,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,   269,     0,     0,     0,     0,     0,   287,
     288,     0,     0,     0,     0,     0,   289,     0,     0,     0,
       0,     0,   270,     0,   290,     0,   291,     0,   292,   293,
     271,   272,     0,     0,     0,     0,   625,     0,   273,   274,
     275,   276,   277,   278,   279,   280,     0,     0,   281,   282,
     283,   284,     0,   285,   286,     0,   269,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,   287,   288,   270,     0,   290,     0,   291,
     289,   292,   293,   271,   272,     0,     0,     0,     0,   626,
       0,   273,   274,   275,   276,   277,   278,   279,   280,     0,
       0,   281,   282,   283,   284,     0,   285,   286,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     269,     0,     0,     0,     0,     0,   287,   288,     0,     0,
       0,     0,     0,   289,     0,     0,     0,     0,     0,   270,
       0,   290,     0,   291,     0,   292,   293,   271,   272,     0,
       0,     0,     0,   627,     0,   273,   274,   275,   276,   277,
     278,   279,   280,     0,     0,   281,   282,   283,   284,     0,
     285,   286,     0,   269,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     287,   288,   270,     0,   290,     0,   291,   289,   292,   293,
     271,   272,     0,     0,     0,     0,   629,     0,   273,   274,
     275,   276,   277,   278,   279,   280,     0,     0,   281,   282,
     283,   284,     0,   285,   286,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,   269,     0,     0,
       0,     0,     0,   287,   288,     0,     0,     0,     0,     0,
     289,     0,     0,     0,     0,     0,   270,     0,   290,     0,
     291,     0,   292,   293,   271,   272,     0,     0,     0,     0,
     630,     0,   273,   274,   275,   276,   277,   278,   279,   280,
       0,     0,   281,   282,   283,   284,     0,   285,   286,     0,
     269,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,   287,   288,   270,
       0,   290,     0,   291,   289,   292,   293,   271,   272,     0,
       0,     0,     0,   631,     0,   273,   274,   275,   276,   277,
     278,   279,   280,     0,     0,   281,   282,   283,   284,     0,
     285,   286,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,   269,     0,     0,     0,     0,     0,
     287,   288,     0,     0,     0,     0,     0,   289,     0,     0,
       0,     0,     0,   270,     0,   290,     0,   291,     0,   292,
     293,   271,   272,     0,     0,     0,     0,   632,     0,   273,
     274,   275,   276,   277,   278,   279,   280,     0,     0,   281,
     282,   283,   284,     0,   285,   286,     0,   269,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,   287,   288,   270,     0,   290,     0,
     291,   289,   292,   293,   271,   272,     0,     0,     0,     0,
     633,     0,   273,   274,   275,   276,   277,   278,   279,   280,
       0,     0,   281,   282,   283,   284,     0,   285,   286,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   269,     0,     0,     0,     0,     0,   287,   288,     0,
       0,     0,     0,     0,   289,     0,     0,     0,     0,     0,
     270,     0,   290,     0,   291,     0,   292,   293,   271,   272,
       0,     0,     0,     0,   636,     0,   273,   274,   275,   276,
     277,   278,   279,   280,     0,     0,   281,   282,   283,   284,
       0,   285,   286,     0,   269,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   287,   288,   270,     0,   290,     0,   291,   289,   292,
     293,   271,   272,     0,     0,     0,     0,   637,     0,   273,
     274,   275,   276,   277,   278,   279,   280,     0,     0,   281,
     282,   283,   284,     0,   285,   286,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   269,     0,
       0,     0,     0,     0,   287,   288,     0,     0,     0,     0,
       0,   289,     0,     0,     0,     0,     0,   270,     0,   290,
       0,   291,     0,   292,   293,   271,   272,     0,     0,     0,
       0,   638,     0,   273,   274,   275,   276,   277,   278,   279,
     280,     0,     0,   281,   282,   283,   284,     0,   285,   286,
       0,   269,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   287,   288,
     270,     0,   290,     0,   291,   289,   292,   293,   271,   272,
       0,     0,     0,     0,   639,     0,   273,   274,   275,   276,
     277,   278,   279,   280,     0,     0,   281,   282,   283,   284,
       0,   285,   286,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,   269,     0,     0,     0,     0,
       0,   287,   288,     0,     0,     0,     0,     0,   289,     0,
       0,     0,     0,     0,   270,     0,   290,     0,   291,     0,
     292,   293,   271,   272,     0,     0,     0,     0,   645,     0,
     273,   274,   275,   276,   277,   278,   279,   280,     0,     0,
     281,   282,   283,   284,     0,   285,   286,     0,   269,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,   287,   288,   270,     0,   290,
       0,   291,   289,   292,   293,   271,   272,     0,     0,     0,
       0,   646,     0,   273,   274,   275,   276,   277,   278,   279,
     280,     0,     0,   281,   282,   283,   284,     0,   285,   286,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   269,     0,     0,     0,     0,     0,   287,   288,
       0,     0,     0,     0,     0,   289,     0,     0,     0,     0,
       0,   270,     0,   290,     0,   291,     0,   292,   293,   271,
     272,     0,     0,     0,     0,   647,     0,   273,   274,   275,
     276,   277,   278,   279,   280,     0,     0,   281,   282,   283,
     284,     0,   285,   286,     0,   269,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   287,   288,   270,     0,   290,     0,   291,   289,
     292,   293,   271,   272,     0,     0,     0,     0,   648,     0,
     273,   274,   275,   276,   277,   278,   279,   280,     0,     0,
     281,   282,   283,   284,     0,   285,   286,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   269,
       0,     0,     0,     0,     0,   287,   288,     0,     0,     0,
       0,     0,   289,     0,     0,     0,     0,     0,   270,     0,
     290,     0,   291,     0,   292,   293,   271,   272,     0,     0,
       0,     0,   653,     0,   273,   274,   275,   276,   277,   278,
     279,   280,     0,     0,   281,   282,   283,   284,     0,   285,
     286,     0,   269,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   287,
     288,   270,     0,   290,     0,   291,   289,   292,   293,   271,
     272,     0,     0,     0,     0,   668,     0,   273,   274,   275,
     276,   277,   278,   279,   280,     0,     0,   281,   282,   283,
     284,     0,   285,   286,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   287,   288,     0,     0,     0,     0,     0,   289,
       0,     0,     0,     0,     0,     0,     0,   290,     0,   291,
       0,   292,   293,     0,     0,     0,     0,     0,     0,   673,
       0,    83,     0,    84,     0,     0,     0,     0,     0,     0,
      85,    86,    87,     0,    88,    89,    90,    91,    92,    93,
       0,     0,     0,     0,     0,     0,     0,    94,     0,     0,
       0,    95,   608,     0,     0,     0,     0,     0,     0,     0,
     290,    96,   291,     0,   292,   293,     0,     0,     0,     0,
       0,    97,   674,     0,    98,    99,   100,   101,   102,   103,
     104,   105,   106,   107,   108,   109,     0,     0,   110,     0,
       0,   111,     0,     0,   112,   113,   114,   115,   116,     0,
     117,   118,   119,   120,   121,   122,   123,   124,   125,   126,
     127,   128,   129,   130,   131,   132,     0,     0,     0,     0,
       0,    83,     0,    84,     0,     0,     0,     0,     0,     0,
      85,    86,    87,     0,    88,    89,    90,    91,    92,    93,
       0,     0,     0,     0,     0,     0,     0,    94,     0,     0,
       0,    95,     0,     0,     0,     0,     0,     0,     0,     0,
       0,    96,     0,     0,     0,   133,    47,     0,   134,   135,
     136,    97,     0,     0,    98,    99,   100,   101,   102,   103,
     104,   105,   106,   107,   108,   109,     0,     0,   110,     0,
       0,   111,     0,     0,   112,   113,   114,   115,   116,     0,
     117,   118,   119,   120,   121,   122,   123,   124,   125,   126,
     127,   128,   129,   130,   131,   132,     0,     0,     0,     0,
       0,    83,     0,    84,     0,     0,     0,     0,     0,     0,
      85,    86,    87,     0,    88,    89,    90,    91,    92,    93,
       0,     0,     0,     0,     0,     0,     0,    94,     0,     0,
       0,    95,     0,     0,     0,     0,     0,     0,     0,     0,
       0,    96,     0,     0,     0,   133,    47,     0,   134,   135,
     136,    97,     0,     0,    98,    99,   100,   101,   102,   103,
     104,   105,   106,   107,   108,   109,     0,     0,   110,     0,
       0,   111,     0,     0,   112,   113,   114,   115,   116,     0,
     117,   118,   119,   120,   121,   122,   123,   124,   125,   126,
     127,   128,   129,   130,   131,   132,     0,     0,     0,     0,
       0,    83,     0,    84,     0,     0,     1,     0,     0,     0,
      85,    86,    87,     0,    88,    89,    90,    91,    92,    93,
       0,     0,     0,     0,     0,     0,     0,    94,     0,     0,
       0,    95,     0,     0,     0,     2,     0,     0,     0,     0,
       0,    96,     0,     0,     0,   401,    47,     0,   134,   135,
     136,    97,     0,     0,    98,    99,   100,   101,   102,   103,
     104,   105,   106,   107,   108,   109,     0,     0,   110,     0,
       0,   111,     0,     0,   112,   113,   114,   115,   116,     0,
     117,   118,   119,   120,   121,   122,   123,   124,   125,   126,
     127,   128,   129,   130,   131,   132,     0,     3,     4,     5,
       6,     7,     8,     9,    10,    11,    12,    13,    14,    15,
      16,    17,    18,    19,    20,    21,    22,    23,    24,    25,
      26,    27,    28,    29,    30,    31,    32,    33,    34,    35,
      36,    37,    38,    39,    40,    41,    42,    43,    44,    45,
      46,   269,     0,     0,     0,   133,    47,    47,   415,   135,
     136,     0,    48,     0,     0,     0,     0,     0,     0,     0,
     270,     0,     0,     0,     0,     0,     0,     0,   271,   272,
       0,     0,     0,     0,     0,     0,   273,   274,   275,   276,
     277,   278,   279,   280,     0,     0,   281,   282,   283,   284,
       0,   285,   286,     0,   269,     0,     0,     0,   550,   551,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   287,   288,   270,     0,     0,     0,     0,   289,     0,
       0,   271,   272,     0,     0,     0,     0,     0,     0,   273,
     274,   275,   276,   277,   278,   279,   280,     0,     0,   281,
     282,   283,   284,     0,   285,   286,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   269,     0,
       0,     0,     0,     0,   287,   288,     0,     0,     0,     0,
       0,   289,     0,     0,     0,     0,     0,   270,     0,   290,
       0,   291,   339,   292,   293,   271,   272,   134,     0,     0,
       0,     0,     0,   273,   274,   275,   276,   277,   278,   279,
     280,     0,     0,   281,   282,   283,   284,     0,   285,   286,
       0,   269,     0,     0,     0,     0,     0,     0,   663,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   287,   288,
     270,     0,   290,     0,   291,   289,   292,   293,   271,   272,
       0,     0,     0,     0,     0,     0,   273,   274,   275,   276,
     277,   278,   279,   280,     0,     0,   281,   282,   283,   284,
       0,   285,   286,     0,     0,     0,     0,   269,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,   287,   288,     0,     0,     0,   270,     0,   289,     0,
       0,   664,     0,     0,   271,   272,   290,     0,   291,     0,
     292,   293,   273,   274,   275,   276,   277,   278,   279,   280,
       0,     0,   281,   282,   283,   284,     0,   285,   286,     0,
     269,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,   287,   288,   270,
       0,     0,     0,     0,   289,     0,     0,   271,   272,   290,
       0,   291,     0,   292,   293,   273,   274,   275,   276,   277,
     278,   279,   280,     0,     0,   281,   282,   283,   284,     0,
     285,   286,     0,     0,     0,     0,   269,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     287,   288,     0,     0,     0,   270,     0,   289,     0,     0,
       0,     0,     0,   271,   272,   290,     0,   291,     0,   292,
     293,   273,   274,   275,   276,   277,   278,   279,   280,     0,
       0,   281,   282,   283,   284,     0,   285,   286,     0,   269,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,   287,   288,   270,     0,
       0,     0,     0,   289,     0,     0,   271,   272,   290,     0,
     291,     0,   292,   293,   273,   274,     0,   276,   277,   278,
     279,   280,     0,     0,   281,   282,   283,   284,     0,   285,
     286,     0,   269,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,   287,
     288,     0,     0,     0,     0,     0,   289,     0,     0,   271,
       0,     0,     0,     0,   290,     0,   654,   273,   292,   293,
     276,   277,   278,   279,   280,     0,     0,   281,   282,   283,
     284,     0,   285,   286,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,   287,   288,     0,     0,     0,     0,     0,   289,
       0,     0,   271,     0,     0,     0,     0,   290,     0,   291,
     273,   292,   293,   276,   277,   278,   279,   280,     0,     0,
     281,   282,   283,   284,     0,   285,   286,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,   287,   288,     0,     0,     0,
       0,     0,   289,     0,     0,   271,     0,     0,     0,     0,
     290,     0,   291,    -1,   292,   293,   276,   277,   278,   279,
     280,     0,     0,   281,   282,   283,   284,     0,   285,   286,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,   287,   288,
       0,     0,     0,     0,     0,   289,     0,     0,     0,     0,
       0,     0,     0,   290,     0,   291,     0,   292,   293,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
       0,     0,     0,     0,     0,     0,    -1,     0,   291,     0,
     292,   293
};

static const yytype_int16 yycheck[] =
{
       4,     5,     6,    82,     8,    10,   206,    43,   215,    71,
      14,    15,    10,    11,    12,    13,    35,    69,   147,   147,
      18,    19,    22,    72,    43,    72,   155,   148,   157,    54,
      55,    33,    34,    35,    36,    37,    38,   149,    42,    43,
      44,    45,    40,    72,   142,    72,    72,   142,   154,     5,
       6,   157,   154,    54,    55,   157,    72,   155,   162,   150,
     155,     5,     6,    19,   156,   157,   170,   171,    24,    25,
      26,    27,    72,    74,    75,    19,     0,     5,     6,    35,
      24,    25,    26,    27,    88,   153,    30,    43,    92,    93,
     153,    19,    96,   142,    31,   157,    24,    25,    26,    27,
     611,   205,   613,    54,    55,   157,   154,   111,    45,   157,
     157,   153,    68,   149,   150,   142,   142,    54,    55,   144,
     142,   146,   147,    74,    68,   154,   142,   147,   157,   133,
     144,   150,   146,   154,   148,   214,   157,    74,    75,   154,
      68,   220,   157,   144,    81,   146,   147,   144,   147,   146,
      72,   148,   157,   144,   147,   146,   147,   159,   147,   216,
     217,   159,   219,   148,   154,   150,   154,   157,   170,   157,
     154,   154,   154,   157,   157,   157,   154,   147,   147,   157,
     147,   147,   186,   147,   147,   147,   147,   147,   144,   147,
     147,   147,   148,   144,   150,   146,   147,   199,   147,   147,
     144,   147,   147,   147,   148,   147,   147,   144,   147,   146,
     147,   315,   147,   317,   293,   319,   144,   145,   147,   147,
     148,   147,   147,   227,   147,   229,   147,   147,   232,   233,
     234,   235,   236,   237,   238,   239,   240,   444,   147,   343,
     147,   147,   246,   247,   248,   249,   250,   251,   252,   253,
     254,   255,   256,   257,   258,   259,   260,   261,   262,   263,
     264,   265,   266,   147,   147,   269,   270,   147,   272,   273,
     274,   275,   276,   277,   278,   279,   280,   281,   282,   283,
     284,   147,   147,   287,   288,   289,   290,   291,   147,   147,
     292,   147,   294,   147,     5,     6,   147,   147,   302,    15,
     155,   303,    31,   150,   157,   150,   308,   150,    19,   157,
      38,   150,   150,    24,    25,    26,    27,   421,   422,    30,
     424,    15,   157,   155,   428,    54,    55,    37,   432,   433,
     334,   435,   157,   155,   157,   339,   536,   154,   342,   443,
     344,   142,   446,   155,   448,    74,    75,     0,     1,    54,
      55,   147,    81,    75,   154,   157,   360,    68,   157,   157,
     157,    38,   155,    16,    17,    30,   154,   142,    72,    74,
      75,   155,   155,   155,   155,   155,    81,   150,   150,   150,
     150,   385,   150,   150,   388,   150,   465,   389,    41,     5,
       6,   150,   156,    46,   156,   155,   157,   401,   154,   154,
     154,   154,   154,    19,   154,   156,    31,   147,    24,    25,
      26,    27,   158,   150,    38,   144,   154,   146,   147,    44,
      45,    46,   155,   606,   211,   442,   530,   450,   171,    54,
      55,   535,   672,   144,   436,   430,   147,   148,    91,   144,
     549,   146,   147,   447,   504,   362,   345,    -1,   425,    74,
      75,    -1,    68,    -1,    -1,    -1,    81,   110,    -1,    -1,
      -1,    -1,   466,   467,   468,   469,   470,   471,   472,   473,
     474,    -1,   476,    -1,    -1,   479,   480,   481,   482,   483,
      -1,    -1,    -1,   487,   488,   489,   490,    -1,    -1,    -1,
      -1,   595,   496,   497,   498,   499,    -1,    -1,    -1,    -1,
     604,   505,   506,    -1,    -1,   507,    -1,    -1,    -1,   162,
      -1,   164,    -1,    -1,    -1,   519,   169,   170,   171,   144,
      -1,   146,   147,    -1,    -1,   527,    -1,   531,   144,    -1,
      -1,   147,   148,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,   547,   548,   549,   550,   551,    -1,   553,
     554,   204,   205,   206,    -1,     4,   209,    -1,   211,    -1,
      -1,    -1,   215,   216,   217,    -1,   219,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    23,    -1,    -1,    -1,    -1,    -1,
     659,    -1,    31,    32,    -1,    -1,    -1,    -1,    -1,   593,
      39,    40,    41,    42,    43,    44,    45,    46,    -1,    -1,
      49,    50,    51,    52,    -1,    54,    55,   611,    -1,   613,
      -1,    -1,    -1,    -1,    -1,   619,    -1,    -1,   271,    -1,
      -1,    -1,    -1,    -1,   628,    74,    75,    -1,    -1,    -1,
      -1,    -1,    81,    -1,    31,    -1,    -1,    -1,    -1,   292,
      -1,    -1,    -1,    -1,    -1,    42,    43,    44,    45,    46,
     654,   655,    49,    50,    51,    52,    -1,    54,    55,    -1,
     664,    -1,   315,    -1,   317,    -1,   319,    -1,    -1,    -1,
      -1,    -1,   676,    -1,    -1,     4,    -1,    74,    75,    -1,
      -1,    -1,    -1,    -1,    81,    -1,    -1,    -1,    -1,    -1,
     343,    -1,   345,   142,    23,   144,    -1,   146,   147,    -1,
      -1,    -1,    31,    32,    -1,   154,    -1,    -1,   157,   362,
      39,    40,    41,    42,    43,    44,    45,    46,    -1,    -1,
      49,    50,    51,    52,    -1,    54,    55,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,   389,    -1,    -1,    -1,
      -1,   394,    -1,    -1,    -1,    74,    75,   144,    -1,   146,
     147,    -1,    81,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,   421,   422,
      -1,   424,   425,    -1,    -1,   428,    -1,   430,    -1,   432,
     433,    -1,   435,    -1,    -1,    -1,    -1,    -1,    -1,   442,
     443,   444,    -1,   446,    -1,   448,   449,   450,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,   142,    -1,   144,    -1,   146,   147,    -1,
       4,    -1,    -1,    -1,    -1,   154,    -1,    -1,   157,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    23,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    31,    32,    -1,
      -1,   504,    -1,    -1,   507,    39,    40,    41,    42,    43,
      44,    45,    46,    -1,    -1,    49,    50,    51,    52,    -1,
      54,    55,    -1,     4,    -1,    -1,    -1,   530,    -1,    -1,
      -1,    -1,   535,   536,    -1,    -1,    -1,    -1,    -1,    -1,
      74,    75,    23,    -1,    -1,    -1,    -1,    81,    -1,    -1,
      31,    32,    -1,    -1,    -1,    -1,    -1,    -1,    39,    40,
      41,    42,    43,    44,    45,    46,    -1,    -1,    49,    50,
      51,    52,    -1,    54,    55,    -1,    -1,    -1,    -1,     4,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,   595,    74,    75,    -1,    -1,    -1,    23,    -1,
      81,   604,    -1,   606,    -1,    -1,    31,    32,   142,    -1,
     144,    -1,   146,   147,    39,    40,    41,    42,    43,    44,
      45,    46,    -1,   157,    49,    50,    51,    52,    -1,    54,
      55,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,    -1,    74,
      75,   654,    -1,    -1,    -1,    -1,    81,    -1,    -1,    -1,
      -1,   142,    23,   144,    -1,   146,   147,    -1,    -1,    -1,
      31,    32,    -1,    -1,    -1,    -1,   157,    -1,    39,    40,
      41,    42,    43,    44,    45,    46,    -1,    -1,    49,    50,
      51,    52,    -1,    54,    55,    -1,    -1,    -1,    -1,     4,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    74,    75,    -1,    -1,   142,    23,   144,
      81,   146,   147,    -1,    -1,    -1,    31,    32,    -1,    -1,
      -1,    -1,   157,    -1,    39,    40,    41,    42,    43,    44,
      45,    46,    -1,    -1,    49,    50,    51,    52,    -1,    54,
      55,    -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,
      75,    23,    -1,    -1,    -1,    -1,    81,    -1,    -1,    31,
      32,   142,    -1,   144,    -1,   146,   147,    39,    40,    41,
      42,    43,    44,    45,    46,    -1,   157,    49,    50,    51,
      52,    -1,    54,    55,    -1,    -1,    -1,    -1,     4,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    74,    75,    -1,    -1,    -1,    23,    -1,    81,
      -1,    -1,    -1,    -1,    -1,    31,    32,   142,    -1,   144,
      -1,   146,   147,    39,    40,    41,    42,    43,    44,    45,
      46,    -1,   157,    49,    50,    51,    52,    -1,    54,    55,
      -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,
      23,    -1,    -1,    -1,    -1,    81,    -1,    -1,    31,    32,
     142,    -1,   144,    -1,   146,   147,    39,    40,    41,    42,
      43,    44,    45,    46,    -1,   157,    49,    50,    51,    52,
      -1,    54,    55,    -1,    -1,    -1,    -1,     4,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    74,    75,    -1,    -1,    -1,    23,    -1,    81,    -1,
      -1,    -1,    -1,    -1,    31,    32,   142,    -1,   144,    -1,
     146,   147,    39,    40,    41,    42,    43,    44,    45,    46,
      -1,   157,    49,    50,    51,    52,    -1,    54,    55,    -1,
       4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    23,
      -1,    -1,    -1,    -1,    81,    -1,    -1,    31,    32,   142,
      -1,   144,    -1,   146,   147,    39,    40,    41,    42,    43,
      44,    45,    46,    -1,   157,    49,    50,    51,    52,    -1,
      54,    55,    -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      74,    75,    -1,    -1,    -1,    23,    -1,    81,    -1,    -1,
      -1,    -1,    -1,    31,    32,   142,    -1,   144,    -1,   146,
     147,    39,    40,    41,    42,    43,    44,    45,    46,    -1,
     157,    49,    50,    51,    52,    -1,    54,    55,    -1,     4,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    23,    -1,
      -1,    -1,    -1,    81,    -1,    -1,    31,    32,   142,    -1,
     144,    -1,   146,   147,    39,    40,    41,    42,    43,    44,
      45,    46,    -1,   157,    49,    50,    51,    52,    -1,    54,
      55,    -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,
      75,    -1,    -1,    -1,    23,    -1,    81,    -1,    -1,    -1,
      -1,    -1,    31,    32,   142,    -1,   144,    -1,   146,   147,
      39,    40,    41,    42,    43,    44,    45,    46,    -1,   157,
      49,    50,    51,    52,    -1,    54,    55,    -1,     4,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    74,    75,    23,    -1,    -1,
      -1,    -1,    81,    -1,    -1,    31,    32,   142,    -1,   144,
      -1,   146,   147,    39,    40,    41,    42,    43,    44,    45,
      46,    -1,   157,    49,    50,    51,    52,    -1,    54,    55,
      -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,
      -1,    -1,    -1,    23,    -1,    81,    -1,    -1,    -1,    -1,
      -1,    31,    32,   142,    -1,   144,    -1,   146,   147,    39,
      40,    41,    42,    43,    44,    45,    46,    -1,   157,    49,
      50,    51,    52,    -1,    54,    55,    -1,     4,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    74,    75,    23,    -1,    -1,    -1,
      -1,    81,    -1,    -1,    31,    32,   142,    -1,   144,    -1,
     146,   147,    39,    40,    41,    42,    43,    44,    45,    46,
      -1,   157,    49,    50,    51,    52,    -1,    54,    55,    -1,
      -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    -1,
      -1,    -1,    23,    -1,    81,    -1,    -1,    -1,    -1,    -1,
      31,    32,   142,    -1,   144,    -1,   146,   147,    39,    40,
      41,    42,    43,    44,    45,    46,    -1,   157,    49,    50,
      51,    52,    -1,    54,    55,    -1,     4,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    74,    75,    23,    -1,    -1,    -1,    -1,
      81,    -1,    -1,    31,    32,   142,    -1,   144,    -1,   146,
     147,    39,    40,    41,    42,    43,    44,    45,    46,    -1,
     157,    49,    50,    51,    52,    -1,    54,    55,    -1,    -1,
      -1,    -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    -1,    -1,
      -1,    23,    -1,    81,    -1,    -1,    -1,    -1,    -1,    31,
      32,   142,    -1,   144,    -1,   146,   147,    39,    40,    41,
      42,    43,    44,    45,    46,    -1,   157,    49,    50,    51,
      52,    -1,    54,    55,    -1,     4,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    74,    75,    23,    -1,    -1,    -1,    -1,    81,
      -1,    -1,    31,    32,   142,    -1,   144,    -1,   146,   147,
      39,    40,    41,    42,    43,    44,    45,    46,    -1,   157,
      49,    50,    51,    52,    -1,    54,    55,    -1,    -1,    -1,
      -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    74,    75,    -1,    -1,    -1,
      23,    -1,    81,    -1,    -1,    -1,    -1,    -1,    31,    32,
     142,    -1,   144,    -1,   146,   147,    39,    40,    41,    42,
      43,    44,    45,    46,    -1,   157,    49,    50,    51,    52,
      -1,    54,    55,    -1,     4,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    74,    75,    23,    -1,    -1,    -1,    -1,    81,    -1,
      -1,    31,    32,   142,    -1,   144,    -1,   146,   147,    39,
      40,    41,    42,    43,    44,    45,    46,    -1,   157,    49,
      50,    51,    52,    -1,    54,    55,    -1,    -1,    -1,    -1,
       4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    74,    75,    -1,    -1,    -1,    23,
      -1,    81,    -1,    -1,    -1,    -1,    -1,    31,    32,   142,
      -1,   144,    -1,   146,   147,    39,    40,    41,    42,    43,
      44,    45,    46,    -1,   157,    49,    50,    51,    52,    -1,
      54,    55,    -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      74,    75,    23,    -1,    -1,    -1,    -1,    81,    -1,    -1,
      31,    32,   142,    -1,   144,    -1,   146,   147,    39,    40,
      41,    42,    43,    44,    45,    46,    -1,   157,    49,    50,
      51,    52,    -1,    54,    55,    -1,    -1,    -1,    -1,     4,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    74,    75,    -1,    -1,    -1,    23,    -1,
      81,    -1,    -1,    -1,    -1,    -1,    31,    32,   142,    -1,
     144,    -1,   146,   147,    39,    40,    41,    42,    43,    44,
      45,    46,    -1,   157,    49,    50,    51,    52,    -1,    54,
      55,    -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,
      75,    23,    -1,    -1,    -1,    -1,    81,    -1,    -1,    31,
      32,   142,    -1,   144,    -1,   146,   147,    39,    40,    41,
      42,    43,    44,    45,    46,    -1,   157,    49,    50,    51,
      52,    -1,    54,    55,    -1,    -1,    -1,    -1,     4,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    74,    75,    -1,    -1,    -1,    23,    -1,    81,
      -1,    -1,    -1,    -1,    -1,    31,    32,   142,    -1,   144,
      -1,   146,   147,    39,    40,    41,    42,    43,    44,    45,
      46,    -1,   157,    49,    50,    51,    52,    -1,    54,    55,
      -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,
      23,    -1,    -1,    -1,    -1,    81,    -1,    -1,    31,    32,
     142,    -1,   144,    -1,   146,   147,    39,    40,    41,    42,
      43,    44,    45,    46,    -1,   157,    49,    50,    51,    52,
      -1,    54,    55,    -1,    -1,    -1,    -1,     4,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    74,    75,    -1,    -1,    -1,    23,    -1,    81,    -1,
      -1,    -1,    -1,    -1,    31,    32,   142,    -1,   144,    -1,
     146,   147,    39,    40,    41,    42,    43,    44,    45,    46,
      -1,   157,    49,    50,    51,    52,    -1,    54,    55,    -1,
       4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    23,
      -1,    -1,    -1,    -1,    81,    -1,    -1,    31,    32,   142,
      -1,   144,    -1,   146,   147,    39,    40,    41,    42,    43,
      44,    45,    46,    -1,   157,    49,    50,    51,    52,    -1,
      54,    55,    -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      74,    75,    -1,    -1,    -1,    23,    -1,    81,    -1,    -1,
      -1,    -1,    -1,    31,    32,   142,    -1,   144,    -1,   146,
     147,    39,    40,    41,    42,    43,    44,    45,    46,    -1,
     157,    49,    50,    51,    52,    -1,    54,    55,    -1,     4,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    23,    -1,
      -1,    -1,    -1,    81,    -1,    -1,    31,    32,   142,    -1,
     144,    -1,   146,   147,    39,    40,    41,    42,    43,    44,
      45,    46,    -1,   157,    49,    50,    51,    52,    -1,    54,
      55,    -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,
      75,    -1,    -1,    -1,    23,    -1,    81,    -1,    -1,    -1,
      -1,    -1,    31,    32,   142,    -1,   144,    -1,   146,   147,
      39,    40,    41,    42,    43,    44,    45,    46,   156,    -1,
      49,    50,    51,    52,    -1,    54,    55,    -1,     4,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    74,    75,    23,    -1,    -1,
      -1,    -1,    81,    -1,    -1,    31,    32,   142,    -1,   144,
      -1,   146,   147,    39,    40,    41,    42,    43,    44,    45,
      46,   156,    -1,    49,    50,    51,    52,    -1,    54,    55,
      -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,
      -1,    -1,    -1,    23,    -1,    81,    -1,    -1,    -1,    -1,
      -1,    31,    32,   142,    -1,   144,    -1,   146,   147,    39,
      40,    41,    42,    43,    44,    45,    46,   156,    -1,    49,
      50,    51,    52,    -1,    54,    55,    -1,     4,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    74,    75,    23,    -1,    -1,    -1,
      -1,    81,    -1,    -1,    31,    32,   142,    -1,   144,    -1,
     146,   147,    39,    40,    41,    42,    43,    44,    45,    46,
     156,    -1,    49,    50,    51,    52,    -1,    54,    55,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,     4,    -1,    -1,    -1,    -1,    -1,    74,    75,    -1,
      -1,    -1,    -1,    -1,    81,    -1,    -1,    -1,    -1,    -1,
      23,    -1,   142,    -1,   144,    -1,   146,   147,    31,    32,
      -1,    -1,    -1,    -1,    -1,   155,    39,    40,    41,    42,
      43,    44,    45,    46,    -1,    -1,    49,    50,    51,    52,
      -1,    54,    55,    -1,     4,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    74,    75,    23,    -1,   142,    -1,   144,    81,   146,
     147,    31,    32,    -1,    -1,    -1,    -1,   154,    -1,    39,
      40,    41,    42,    43,    44,    45,    46,    -1,    -1,    49,
      50,    51,    52,    -1,    54,    55,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,     4,    -1,
      -1,    -1,    -1,    -1,    74,    75,    -1,    -1,    -1,    -1,
      -1,    81,    -1,    -1,    -1,    -1,    -1,    23,    -1,   142,
      -1,   144,    -1,   146,   147,    31,    32,    -1,    -1,    -1,
      -1,   154,    -1,    39,    40,    41,    42,    43,    44,    45,
      46,    -1,    -1,    49,    50,    51,    52,    -1,    54,    55,
      -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,
      23,    -1,   142,    -1,   144,    81,   146,   147,    31,    32,
      -1,    -1,    -1,    -1,   154,    -1,    39,    40,    41,    42,
      43,    44,    45,    46,    -1,    -1,    49,    50,    51,    52,
      -1,    54,    55,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,
      -1,    74,    75,    -1,    -1,    -1,    -1,    -1,    81,    -1,
      -1,    -1,    -1,    -1,    23,    -1,   142,    -1,   144,    -1,
     146,   147,    31,    32,    -1,    -1,    -1,    -1,   154,    -1,
      39,    40,    41,    42,    43,    44,    45,    46,    -1,    -1,
      49,    50,    51,    52,    -1,    54,    55,    -1,     4,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    74,    75,    23,    -1,   142,
      -1,   144,    81,   146,   147,    31,    32,    -1,    -1,    -1,
      -1,   154,    -1,    39,    40,    41,    42,    43,    44,    45,
      46,    -1,    -1,    49,    50,    51,    52,    -1,    54,    55,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,     4,    -1,    -1,    -1,    -1,    -1,    74,    75,
      -1,    -1,    -1,    -1,    -1,    81,    -1,    -1,    -1,    -1,
      -1,    23,    -1,   142,    -1,   144,    -1,   146,   147,    31,
      32,    -1,    -1,    -1,    -1,   154,    -1,    39,    40,    41,
      42,    43,    44,    45,    46,    -1,    -1,    49,    50,    51,
      52,    -1,    54,    55,    -1,     4,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    74,    75,    23,    -1,   142,    -1,   144,    81,
     146,   147,    31,    32,    -1,    -1,    -1,    -1,   154,    -1,
      39,    40,    41,    42,    43,    44,    45,    46,    -1,    -1,
      49,    50,    51,    52,    -1,    54,    55,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,     4,
      -1,    -1,    -1,    -1,    -1,    74,    75,    -1,    -1,    -1,
      -1,    -1,    81,    -1,    -1,    -1,    -1,    -1,    23,    -1,
     142,    -1,   144,    -1,   146,   147,    31,    32,    -1,    -1,
      -1,    -1,   154,    -1,    39,    40,    41,    42,    43,    44,
      45,    46,    -1,    -1,    49,    50,    51,    52,    -1,    54,
      55,    -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,
      75,    23,    -1,   142,    -1,   144,    81,   146,   147,    31,
      32,    -1,    -1,    -1,    -1,   154,    -1,    39,    40,    41,
      42,    43,    44,    45,    46,    -1,    -1,    49,    50,    51,
      52,    -1,    54,    55,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,
      -1,    -1,    74,    75,    -1,    -1,    -1,    -1,    -1,    81,
      -1,    -1,    -1,    -1,    -1,    23,    -1,   142,    -1,   144,
      -1,   146,   147,    31,    32,    -1,    -1,    -1,    -1,   154,
      -1,    39,    40,    41,    42,    43,    44,    45,    46,    -1,
      -1,    49,    50,    51,    52,    -1,    54,    55,    -1,     4,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    23,    -1,
     142,    -1,   144,    81,   146,   147,    31,    32,    -1,    -1,
      -1,    -1,   154,    -1,    39,    40,    41,    42,    43,    44,
      45,    46,    -1,    -1,    49,    50,    51,    52,    -1,    54,
      55,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,    -1,    74,
      75,    -1,    -1,    -1,    -1,    -1,    81,    -1,    -1,    -1,
      -1,    -1,    23,    -1,   142,    -1,   144,    -1,   146,   147,
      31,    32,    -1,    -1,    -1,    -1,   154,    -1,    39,    40,
      41,    42,    43,    44,    45,    46,    -1,    -1,    49,    50,
      51,    52,    -1,    54,    55,    -1,     4,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    74,    75,    23,    -1,   142,    -1,   144,
      81,   146,   147,    31,    32,    -1,    -1,    -1,    -1,   154,
      -1,    39,    40,    41,    42,    43,    44,    45,    46,    -1,
      -1,    49,    50,    51,    52,    -1,    54,    55,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
       4,    -1,    -1,    -1,    -1,    -1,    74,    75,    -1,    -1,
      -1,    -1,    -1,    81,    -1,    -1,    -1,    -1,    -1,    23,
      -1,   142,    -1,   144,    -1,   146,   147,    31,    32,    -1,
      -1,    -1,    -1,   154,    -1,    39,    40,    41,    42,    43,
      44,    45,    46,    -1,    -1,    49,    50,    51,    52,    -1,
      54,    55,    -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      74,    75,    23,    -1,   142,    -1,   144,    81,   146,   147,
      31,    32,    -1,    -1,    -1,    -1,   154,    -1,    39,    40,
      41,    42,    43,    44,    45,    46,    -1,    -1,    49,    50,
      51,    52,    -1,    54,    55,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,     4,    -1,    -1,
      -1,    -1,    -1,    74,    75,    -1,    -1,    -1,    -1,    -1,
      81,    -1,    -1,    -1,    -1,    -1,    23,    -1,   142,    -1,
     144,    -1,   146,   147,    31,    32,    -1,    -1,    -1,    -1,
     154,    -1,    39,    40,    41,    42,    43,    44,    45,    46,
      -1,    -1,    49,    50,    51,    52,    -1,    54,    55,    -1,
       4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    23,
      -1,   142,    -1,   144,    81,   146,   147,    31,    32,    -1,
      -1,    -1,    -1,   154,    -1,    39,    40,    41,    42,    43,
      44,    45,    46,    -1,    -1,    49,    50,    51,    52,    -1,
      54,    55,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,    -1,
      74,    75,    -1,    -1,    -1,    -1,    -1,    81,    -1,    -1,
      -1,    -1,    -1,    23,    -1,   142,    -1,   144,    -1,   146,
     147,    31,    32,    -1,    -1,    -1,    -1,   154,    -1,    39,
      40,    41,    42,    43,    44,    45,    46,    -1,    -1,    49,
      50,    51,    52,    -1,    54,    55,    -1,     4,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    74,    75,    23,    -1,   142,    -1,
     144,    81,   146,   147,    31,    32,    -1,    -1,    -1,    -1,
     154,    -1,    39,    40,    41,    42,    43,    44,    45,    46,
      -1,    -1,    49,    50,    51,    52,    -1,    54,    55,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,     4,    -1,    -1,    -1,    -1,    -1,    74,    75,    -1,
      -1,    -1,    -1,    -1,    81,    -1,    -1,    -1,    -1,    -1,
      23,    -1,   142,    -1,   144,    -1,   146,   147,    31,    32,
      -1,    -1,    -1,    -1,   154,    -1,    39,    40,    41,    42,
      43,    44,    45,    46,    -1,    -1,    49,    50,    51,    52,
      -1,    54,    55,    -1,     4,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    74,    75,    23,    -1,   142,    -1,   144,    81,   146,
     147,    31,    32,    -1,    -1,    -1,    -1,   154,    -1,    39,
      40,    41,    42,    43,    44,    45,    46,    -1,    -1,    49,
      50,    51,    52,    -1,    54,    55,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,     4,    -1,
      -1,    -1,    -1,    -1,    74,    75,    -1,    -1,    -1,    -1,
      -1,    81,    -1,    -1,    -1,    -1,    -1,    23,    -1,   142,
      -1,   144,    -1,   146,   147,    31,    32,    -1,    -1,    -1,
      -1,   154,    -1,    39,    40,    41,    42,    43,    44,    45,
      46,    -1,    -1,    49,    50,    51,    52,    -1,    54,    55,
      -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,
      23,    -1,   142,    -1,   144,    81,   146,   147,    31,    32,
      -1,    -1,    -1,    -1,   154,    -1,    39,    40,    41,    42,
      43,    44,    45,    46,    -1,    -1,    49,    50,    51,    52,
      -1,    54,    55,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,    -1,
      -1,    74,    75,    -1,    -1,    -1,    -1,    -1,    81,    -1,
      -1,    -1,    -1,    -1,    23,    -1,   142,    -1,   144,    -1,
     146,   147,    31,    32,    -1,    -1,    -1,    -1,   154,    -1,
      39,    40,    41,    42,    43,    44,    45,    46,    -1,    -1,
      49,    50,    51,    52,    -1,    54,    55,    -1,     4,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    74,    75,    23,    -1,   142,
      -1,   144,    81,   146,   147,    31,    32,    -1,    -1,    -1,
      -1,   154,    -1,    39,    40,    41,    42,    43,    44,    45,
      46,    -1,    -1,    49,    50,    51,    52,    -1,    54,    55,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,     4,    -1,    -1,    -1,    -1,    -1,    74,    75,
      -1,    -1,    -1,    -1,    -1,    81,    -1,    -1,    -1,    -1,
      -1,    23,    -1,   142,    -1,   144,    -1,   146,   147,    31,
      32,    -1,    -1,    -1,    -1,   154,    -1,    39,    40,    41,
      42,    43,    44,    45,    46,    -1,    -1,    49,    50,    51,
      52,    -1,    54,    55,    -1,     4,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    74,    75,    23,    -1,   142,    -1,   144,    81,
     146,   147,    31,    32,    -1,    -1,    -1,    -1,   154,    -1,
      39,    40,    41,    42,    43,    44,    45,    46,    -1,    -1,
      49,    50,    51,    52,    -1,    54,    55,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,     4,
      -1,    -1,    -1,    -1,    -1,    74,    75,    -1,    -1,    -1,
      -1,    -1,    81,    -1,    -1,    -1,    -1,    -1,    23,    -1,
     142,    -1,   144,    -1,   146,   147,    31,    32,    -1,    -1,
      -1,    -1,   154,    -1,    39,    40,    41,    42,    43,    44,
      45,    46,    -1,    -1,    49,    50,    51,    52,    -1,    54,
      55,    -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,
      75,    23,    -1,   142,    -1,   144,    81,   146,   147,    31,
      32,    -1,    -1,    -1,    -1,   154,    -1,    39,    40,    41,
      42,    43,    44,    45,    46,    -1,    -1,    49,    50,    51,
      52,    -1,    54,    55,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    74,    75,    -1,    -1,    -1,    -1,    -1,    81,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,   142,    -1,   144,
      -1,   146,   147,    -1,    -1,    -1,    -1,    -1,    -1,   154,
      -1,     3,    -1,     5,    -1,    -1,    -1,    -1,    -1,    -1,
      12,    13,    14,    -1,    16,    17,    18,    19,    20,    21,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    29,    -1,    -1,
      -1,    33,    34,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
     142,    43,   144,    -1,   146,   147,    -1,    -1,    -1,    -1,
      -1,    53,   154,    -1,    56,    57,    58,    59,    60,    61,
      62,    63,    64,    65,    66,    67,    -1,    -1,    70,    -1,
      -1,    73,    -1,    -1,    76,    77,    78,    79,    80,    -1,
      82,    83,    84,    85,    86,    87,    88,    89,    90,    91,
      92,    93,    94,    95,    96,    97,    -1,    -1,    -1,    -1,
      -1,     3,    -1,     5,    -1,    -1,    -1,    -1,    -1,    -1,
      12,    13,    14,    -1,    16,    17,    18,    19,    20,    21,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    29,    -1,    -1,
      -1,    33,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    43,    -1,    -1,    -1,   147,   148,    -1,   150,   151,
     152,    53,    -1,    -1,    56,    57,    58,    59,    60,    61,
      62,    63,    64,    65,    66,    67,    -1,    -1,    70,    -1,
      -1,    73,    -1,    -1,    76,    77,    78,    79,    80,    -1,
      82,    83,    84,    85,    86,    87,    88,    89,    90,    91,
      92,    93,    94,    95,    96,    97,    -1,    -1,    -1,    -1,
      -1,     3,    -1,     5,    -1,    -1,    -1,    -1,    -1,    -1,
      12,    13,    14,    -1,    16,    17,    18,    19,    20,    21,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    29,    -1,    -1,
      -1,    33,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    43,    -1,    -1,    -1,   147,   148,    -1,   150,   151,
     152,    53,    -1,    -1,    56,    57,    58,    59,    60,    61,
      62,    63,    64,    65,    66,    67,    -1,    -1,    70,    -1,
      -1,    73,    -1,    -1,    76,    77,    78,    79,    80,    -1,
      82,    83,    84,    85,    86,    87,    88,    89,    90,    91,
      92,    93,    94,    95,    96,    97,    -1,    -1,    -1,    -1,
      -1,     3,    -1,     5,    -1,    -1,     7,    -1,    -1,    -1,
      12,    13,    14,    -1,    16,    17,    18,    19,    20,    21,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    29,    -1,    -1,
      -1,    33,    -1,    -1,    -1,    36,    -1,    -1,    -1,    -1,
      -1,    43,    -1,    -1,    -1,   147,   148,    -1,   150,   151,
     152,    53,    -1,    -1,    56,    57,    58,    59,    60,    61,
      62,    63,    64,    65,    66,    67,    -1,    -1,    70,    -1,
      -1,    73,    -1,    -1,    76,    77,    78,    79,    80,    -1,
      82,    83,    84,    85,    86,    87,    88,    89,    90,    91,
      92,    93,    94,    95,    96,    97,    -1,    98,    99,   100,
     101,   102,   103,   104,   105,   106,   107,   108,   109,   110,
     111,   112,   113,   114,   115,   116,   117,   118,   119,   120,
     121,   122,   123,   124,   125,   126,   127,   128,   129,   130,
     131,   132,   133,   134,   135,   136,   137,   138,   139,   140,
     141,     4,    -1,    -1,    -1,   147,   148,   148,   150,   151,
     152,    -1,   153,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      23,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    31,    32,
      -1,    -1,    -1,    -1,    -1,    -1,    39,    40,    41,    42,
      43,    44,    45,    46,    -1,    -1,    49,    50,    51,    52,
      -1,    54,    55,    -1,     4,    -1,    -1,    -1,     8,     9,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    74,    75,    23,    -1,    -1,    -1,    -1,    81,    -1,
      -1,    31,    32,    -1,    -1,    -1,    -1,    -1,    -1,    39,
      40,    41,    42,    43,    44,    45,    46,    -1,    -1,    49,
      50,    51,    52,    -1,    54,    55,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,     4,    -1,
      -1,    -1,    -1,    -1,    74,    75,    -1,    -1,    -1,    -1,
      -1,    81,    -1,    -1,    -1,    -1,    -1,    23,    -1,   142,
      -1,   144,    28,   146,   147,    31,    32,   150,    -1,    -1,
      -1,    -1,    -1,    39,    40,    41,    42,    43,    44,    45,
      46,    -1,    -1,    49,    50,    51,    52,    -1,    54,    55,
      -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    11,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,
      23,    -1,   142,    -1,   144,    81,   146,   147,    31,    32,
      -1,    -1,    -1,    -1,    -1,    -1,    39,    40,    41,    42,
      43,    44,    45,    46,    -1,    -1,    49,    50,    51,    52,
      -1,    54,    55,    -1,    -1,    -1,    -1,     4,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    74,    75,    -1,    -1,    -1,    23,    -1,    81,    -1,
      -1,    28,    -1,    -1,    31,    32,   142,    -1,   144,    -1,
     146,   147,    39,    40,    41,    42,    43,    44,    45,    46,
      -1,    -1,    49,    50,    51,    52,    -1,    54,    55,    -1,
       4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    23,
      -1,    -1,    -1,    -1,    81,    -1,    -1,    31,    32,   142,
      -1,   144,    -1,   146,   147,    39,    40,    41,    42,    43,
      44,    45,    46,    -1,    -1,    49,    50,    51,    52,    -1,
      54,    55,    -1,    -1,    -1,    -1,     4,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      74,    75,    -1,    -1,    -1,    23,    -1,    81,    -1,    -1,
      -1,    -1,    -1,    31,    32,   142,    -1,   144,    -1,   146,
     147,    39,    40,    41,    42,    43,    44,    45,    46,    -1,
      -1,    49,    50,    51,    52,    -1,    54,    55,    -1,     4,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    74,    75,    23,    -1,
      -1,    -1,    -1,    81,    -1,    -1,    31,    32,   142,    -1,
     144,    -1,   146,   147,    39,    40,    -1,    42,    43,    44,
      45,    46,    -1,    -1,    49,    50,    51,    52,    -1,    54,
      55,    -1,     4,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,
      75,    -1,    -1,    -1,    -1,    -1,    81,    -1,    -1,    31,
      -1,    -1,    -1,    -1,   142,    -1,   144,    39,   146,   147,
      42,    43,    44,    45,    46,    -1,    -1,    49,    50,    51,
      52,    -1,    54,    55,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    74,    75,    -1,    -1,    -1,    -1,    -1,    81,
      -1,    -1,    31,    -1,    -1,    -1,    -1,   142,    -1,   144,
      39,   146,   147,    42,    43,    44,    45,    46,    -1,    -1,
      49,    50,    51,    52,    -1,    54,    55,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    74,    75,    -1,    -1,    -1,
      -1,    -1,    81,    -1,    -1,    31,    -1,    -1,    -1,    -1,
     142,    -1,   144,    39,   146,   147,    42,    43,    44,    45,
      46,    -1,    -1,    49,    50,    51,    52,    -1,    54,    55,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    74,    75,
      -1,    -1,    -1,    -1,    -1,    81,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,   142,    -1,   144,    -1,   146,   147,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
      -1,    -1,    -1,    -1,    -1,    -1,   142,    -1,   144,    -1,
     146,   147
};

  /* YYSTOS[STATE-NUM] -- The (internal number of the) accessing
     symbol of state STATE-NUM.  */
static const yytype_uint8 yystos[] =
{
       0,     7,    36,    98,    99,   100,   101,   102,   103,   104,
     105,   106,   107,   108,   109,   110,   111,   112,   113,   114,
     115,   116,   117,   118,   119,   120,   121,   122,   123,   124,
     125,   126,   127,   128,   129,   130,   131,   132,   133,   134,
     135,   136,   137,   138,   139,   140,   141,   148,   153,   160,
     161,   162,   163,   164,   165,   166,   167,   168,   169,   170,
     171,   172,   173,   174,   175,   176,   177,   178,   179,   180,
     181,   182,   183,   184,   185,   186,   194,   249,   250,   186,
     195,   196,   147,     3,     5,    12,    13,    14,    16,    17,
      18,    19,    20,    21,    29,    33,    43,    53,    56,    57,
      58,    59,    60,    61,    62,    63,    64,    65,    66,    67,
      70,    73,    76,    77,    78,    79,    80,    82,    83,    84,
      85,    86,    87,    88,    89,    90,    91,    92,    93,    94,
      95,    96,    97,   147,   150,   151,   152,   186,   188,   189,
     190,   219,   220,   221,   222,   231,   232,   233,   234,   237,
     241,   219,   219,   219,   149,   187,   187,   187,   187,   187,
     219,   219,     5,     6,    19,    24,    25,    26,    27,    68,
     144,   147,   186,   191,   192,   201,   202,   205,   207,   209,
     210,   211,   212,   213,   214,   245,   186,   187,   187,   188,
     188,   188,   188,   188,   188,   187,   186,   219,   219,   219,
     219,   186,     0,   153,   147,   155,   157,   153,   153,   142,
      10,   157,   219,   224,   147,   147,   147,   147,   219,   147,
     147,   186,   238,   239,   240,   219,   219,   147,   219,   147,
     147,   147,   147,   147,   147,   147,   147,   147,   147,   147,
     147,   147,   186,   235,   236,   219,   147,   147,   147,   147,
     147,   147,   147,   147,   147,   147,   147,   147,   147,   147,
     147,   147,   147,   147,   147,   147,   147,   219,   224,     4,
      23,    31,    32,    39,    40,    41,    42,    43,    44,    45,
      46,    49,    50,    51,    52,    54,    55,    74,    75,    81,
     142,   144,   146,   147,    43,   187,   188,   191,   186,   242,
     243,   244,   147,   147,   186,   203,   204,    35,    43,   188,
     191,   206,   215,   191,   206,    72,   219,   155,   188,   155,
     186,   217,   218,   246,   247,   248,    30,   191,   218,   186,
     197,   198,   196,   154,   157,   224,   246,   247,   247,    28,
     247,   224,   142,   155,    15,   157,   219,   219,   150,   150,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   150,
      38,    71,   157,   219,   219,   219,   219,   219,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   219,   219,   219,   154,   157,   219,   219,   144,   146,
     186,   227,   228,   229,   230,   219,   219,   219,   219,   219,
     219,   147,   219,   219,   219,   219,   219,   219,   219,   150,
     150,   219,   219,   219,   219,   150,   219,   186,   188,   224,
     188,    22,   142,   155,    15,   157,   219,   188,   155,    69,
     157,   188,    72,   157,   156,   157,    37,   154,   154,   191,
     191,   191,   157,   155,   157,   154,   142,   142,   155,   147,
      75,   219,   154,   154,   154,   154,   219,   154,   154,   219,
     191,   219,   240,   154,   154,   157,   157,   157,   157,   157,
     157,   157,   157,   157,   157,   154,   157,   219,   236,   157,
     157,   157,   157,   157,   154,   157,   157,   157,   157,   157,
     157,   157,   157,   157,   157,   157,   157,   157,   157,   157,
     219,   219,   186,   188,   157,    38,   144,   146,   186,   219,
     155,   156,   154,   191,   191,    30,   191,   244,   154,   157,
     154,   191,   204,   191,   191,   191,    35,    43,   188,   216,
      72,   142,   217,   191,   246,   155,   145,   191,   193,   208,
     219,    30,   191,   186,   199,   200,   198,   155,   155,   155,
       8,     9,   223,   155,   142,   224,   219,   219,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   150,   150,   219,   219,   219,   219,   150,   150,   150,
     150,   150,   219,   219,   219,   219,   154,   156,   227,   219,
     219,   186,   188,   157,   150,   142,   219,   156,   188,   156,
     191,   219,   191,   218,   155,   154,   157,   219,    34,   219,
     225,   226,   219,   226,   219,   219,   219,   219,   154,   157,
     154,   154,   154,   154,   154,   154,   154,   154,   157,   154,
     154,   154,   154,   154,   157,   154,   154,   154,   154,   154,
     154,   154,   154,   154,   154,   154,   154,   154,   154,   156,
     219,   156,   191,   154,   144,   142,   158,   191,   200,   147,
     219,   225,   219,    11,    28,   219,   219,   150,   154,   186,
     219,   224,   219,   154,   154,   154,    38,   154,   223,   219,
     155,   156
};

  /* YYR1[YYN] -- Symbol number of symbol that rule YYN derives.  */
static const yytype_uint8 yyr1[] =
{
       0,   159,   160,   160,   160,   160,   160,   161,   161,   161,
     161,   161,   161,   161,   161,   161,   161,   161,   161,   161,
     161,   161,   161,   161,   161,   161,   161,   161,   161,   161,
     161,   162,   163,   164,   164,   164,   164,   164,   165,   165,
     166,   166,   167,   167,   167,   168,   168,   169,   170,   170,
     171,   172,   172,   173,   174,   174,   174,   174,   174,   174,
     174,   174,   175,   175,   175,   175,   175,   176,   176,   176,
     176,   177,   177,   177,   177,   178,   178,   178,   178,   178,
     179,   179,   180,   181,   182,   183,   184,   185,   186,   187,
     188,   189,   190,   191,   191,   192,   192,   192,   192,   192,
     192,   192,   192,   192,   192,   193,   193,   194,   195,   195,
     196,   197,   197,   198,   198,   199,   199,   200,   201,   201,
     201,   202,   203,   203,   204,   205,   206,   206,   207,   208,
     209,   209,   210,   210,   210,   211,   212,   213,   214,   215,
     215,   215,   216,   216,   216,   217,   217,   218,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   219,   219,   219,   219,   219,   219,   219,   219,   219,
     219,   220,   220,   221,   221,   222,   223,   223,   224,   224,
     225,   226,   226,   227,   228,   228,   229,   230,   230,   230,
     230,   230,   230,   230,   230,   231,   232,   232,   232,   232,
     233,   234,   235,   235,   236,   237,   238,   239,   239,   240,
     240,   241,   242,   243,   243,   244,   244,   245,   246,   247,
     247,   248,   249,   249,   249,   249,   249,   250,   250,   250,
     250
};

  /* YYR2[YYN] -- Number of symbols on the right hand side of rule YYN.  */
static const yytype_uint8 yyr2[] =
{
       0,     2,     2,     2,     2,     1,     1,     1,     1,     1,
       1,     1,     1,     1,     1,     1,     1,     1,     1,     1,
       1,     1,     1,     1,     1,     1,     1,     1,     1,     1,
       1,     4,     2,     2,     2,     1,     1,     2,     2,     1,
       2,     2,     3,     4,     3,     2,     1,     2,     2,     2,
       3,     2,     1,     2,     1,     1,     1,     1,     1,     1,
       1,     1,     1,     1,     1,     1,     1,     2,     1,     2,
       1,     2,     1,     2,     1,     2,     1,     2,     1,     1,
       2,     1,     2,     2,     4,     3,     2,    11,     1,     1,
       1,     1,     1,     1,     1,     1,     1,     1,     1,     1,
       1,     1,     1,     1,     3,     1,     1,     3,     1,     3,
       3,     1,     3,     1,     4,     1,     3,     3,     5,     3,
       5,     3,     1,     3,     3,     3,     3,     3,     4,     3,
       4,     6,     1,     1,     1,     4,     1,     1,     5,     1,
       1,     2,     1,     1,     2,     1,     3,     1,     1,     1,
       1,     1,     4,     4,     4,     3,     3,     3,     1,     1,
       1,     1,     1,     1,     1,     1,     1,     2,     2,     2,
       4,     3,     3,     3,     1,     1,     3,     3,     3,     3,
       7,     3,     3,     3,     3,     3,     4,     3,     3,     3,
       2,     6,     3,     3,     3,     3,     6,     6,     6,     6,
       6,     6,     6,     6,     6,     6,     6,     6,     6,     6,
       6,     6,     6,     6,     3,     8,     4,     6,     6,     8,
       6,     6,     6,     6,     6,     6,     6,     6,     4,     8,
       4,     3,     3,     3,     3,     5,     3,     5,     1,     3,
       5,     1,     2,     3,     1,     3,     1,     3,     1,     2,
       2,     4,     2,     3,     3,     6,     6,     7,     6,     7,
       6,     3,     1,     3,     3,     5,     1,     1,     3,     3,
       5,     4,     1,     1,     3,     3,     5,     4,     3,     1,
       3,     1,     5,     3,     8,     6,     5,     1,     5,     3,
       5
};


#define yyerrok         (yyerrstatus = 0)
#define yyclearin       (yychar = YYEMPTY)
#define YYEMPTY         (-2)
#define YYEOF           0

#define YYACCEPT        goto yyacceptlab
#define YYABORT         goto yyabortlab
#define YYERROR         goto yyerrorlab


#define YYRECOVERING()  (!!yyerrstatus)

#define YYBACKUP(Token, Value)                                  \
do                                                              \
  if (yychar == YYEMPTY)                                        \
    {                                                           \
      yychar = (Token);                                         \
      yylval = (Value);                                         \
      YYPOPSTACK (yylen);                                       \
      yystate = *yyssp;                                         \
      goto yybackup;                                            \
    }                                                           \
  else                                                          \
    {                                                           \
      yyerror (YY_("syntax error: cannot back up")); \
      YYERROR;                                                  \
    }                                                           \
while (0)

/* Error token number */
#define YYTERROR        1
#define YYERRCODE       256



/* Enable debugging if requested.  */
#if YYDEBUG

# ifndef YYFPRINTF
#  include <stdio.h> /* INFRINGES ON USER NAME SPACE */
#  define YYFPRINTF fprintf
# endif

# define YYDPRINTF(Args)                        \
do {                                            \
  if (yydebug)                                  \
    YYFPRINTF Args;                             \
} while (0)

/* This macro is provided for backward compatibility. */
#ifndef YY_LOCATION_PRINT
# define YY_LOCATION_PRINT(File, Loc) ((void) 0)
#endif


# define YY_SYMBOL_PRINT(Title, Type, Value, Location)                    \
do {                                                                      \
  if (yydebug)                                                            \
    {                                                                     \
      YYFPRINTF (stderr, "%s ", Title);                                   \
      yy_symbol_print (stderr,                                            \
                  Type, Value); \
      YYFPRINTF (stderr, "\n");                                           \
    }                                                                     \
} while (0)


/*----------------------------------------.
| Print this symbol's value on YYOUTPUT.  |
`----------------------------------------*/

static void
yy_symbol_value_print (FILE *yyoutput, int yytype, YYSTYPE const * const yyvaluep)
{
  FILE *yyo = yyoutput;
  YYUSE (yyo);
  if (!yyvaluep)
    return;
# ifdef YYPRINT
  if (yytype < YYNTOKENS)
    YYPRINT (yyoutput, yytoknum[yytype], *yyvaluep);
# endif
  YYUSE (yytype);
}


/*--------------------------------.
| Print this symbol on YYOUTPUT.  |
`--------------------------------*/

static void
yy_symbol_print (FILE *yyoutput, int yytype, YYSTYPE const * const yyvaluep)
{
  YYFPRINTF (yyoutput, "%s %s (",
             yytype < YYNTOKENS ? "token" : "nterm", yytname[yytype]);

  yy_symbol_value_print (yyoutput, yytype, yyvaluep);
  YYFPRINTF (yyoutput, ")");
}

/*------------------------------------------------------------------.
| yy_stack_print -- Print the state stack from its BOTTOM up to its |
| TOP (included).                                                   |
`------------------------------------------------------------------*/

static void
yy_stack_print (yytype_int16 *yybottom, yytype_int16 *yytop)
{
  YYFPRINTF (stderr, "Stack now");
  for (; yybottom <= yytop; yybottom++)
    {
      int yybot = *yybottom;
      YYFPRINTF (stderr, " %d", yybot);
    }
  YYFPRINTF (stderr, "\n");
}

# define YY_STACK_PRINT(Bottom, Top)                            \
do {                                                            \
  if (yydebug)                                                  \
    yy_stack_print ((Bottom), (Top));                           \
} while (0)


/*------------------------------------------------.
| Report that the YYRULE is going to be reduced.  |
`------------------------------------------------*/

static void
yy_reduce_print (yytype_int16 *yyssp, YYSTYPE *yyvsp, int yyrule)
{
  unsigned long int yylno = yyrline[yyrule];
  int yynrhs = yyr2[yyrule];
  int yyi;
  YYFPRINTF (stderr, "Reducing stack by rule %d (line %lu):\n",
             yyrule - 1, yylno);
  /* The symbols being reduced.  */
  for (yyi = 0; yyi < yynrhs; yyi++)
    {
      YYFPRINTF (stderr, "   $%d = ", yyi + 1);
      yy_symbol_print (stderr,
                       yystos[yyssp[yyi + 1 - yynrhs]],
                       &(yyvsp[(yyi + 1) - (yynrhs)])
                                              );
      YYFPRINTF (stderr, "\n");
    }
}

# define YY_REDUCE_PRINT(Rule)          \
do {                                    \
  if (yydebug)                          \
    yy_reduce_print (yyssp, yyvsp, Rule); \
} while (0)

/* Nonzero means print parse trace.  It is left uninitialized so that
   multiple parsers can coexist.  */
int yydebug;
#else /* !YYDEBUG */
# define YYDPRINTF(Args)
# define YY_SYMBOL_PRINT(Title, Type, Value, Location)
# define YY_STACK_PRINT(Bottom, Top)
# define YY_REDUCE_PRINT(Rule)
#endif /* !YYDEBUG */


/* YYINITDEPTH -- initial size of the parser's stacks.  */
#ifndef YYINITDEPTH
# define YYINITDEPTH 200
#endif

/* YYMAXDEPTH -- maximum size the stacks can grow to (effective only
   if the built-in stack extension method is used).

   Do not make this value too large; the results are undefined if
   YYSTACK_ALLOC_MAXIMUM < YYSTACK_BYTES (YYMAXDEPTH)
   evaluated with infinite-precision integer arithmetic.  */

#ifndef YYMAXDEPTH
# define YYMAXDEPTH 10000
#endif


#if YYERROR_VERBOSE

# ifndef yystrlen
#  if defined __GLIBC__ && defined _STRING_H
#   define yystrlen strlen
#  else
/* Return the length of YYSTR.  */
static YYSIZE_T
yystrlen (const char *yystr)
{
  YYSIZE_T yylen;
  for (yylen = 0; yystr[yylen]; yylen++)
    continue;
  return yylen;
}
#  endif
# endif

# ifndef yystpcpy
#  if defined __GLIBC__ && defined _STRING_H && defined _GNU_SOURCE
#   define yystpcpy stpcpy
#  else
/* Copy YYSRC to YYDEST, returning the address of the terminating '\0' in
   YYDEST.  */
static char *
yystpcpy (char *yydest, const char *yysrc)
{
  char *yyd = yydest;
  const char *yys = yysrc;

  while ((*yyd++ = *yys++) != '\0')
    continue;

  return yyd - 1;
}
#  endif
# endif

# ifndef yytnamerr
/* Copy to YYRES the contents of YYSTR after stripping away unnecessary
   quotes and backslashes, so that it's suitable for yyerror.  The
   heuristic is that double-quoting is unnecessary unless the string
   contains an apostrophe, a comma, or backslash (other than
   backslash-backslash).  YYSTR is taken from yytname.  If YYRES is
   null, do not copy; instead, return the length of what the result
   would have been.  */
static YYSIZE_T
yytnamerr (char *yyres, const char *yystr)
{
  if (*yystr == '"')
    {
      YYSIZE_T yyn = 0;
      char const *yyp = yystr;

      for (;;)
        switch (*++yyp)
          {
          case '\'':
          case ',':
            goto do_not_strip_quotes;

          case '\\':
            if (*++yyp != '\\')
              goto do_not_strip_quotes;
            /* Fall through.  */
          default:
            if (yyres)
              yyres[yyn] = *yyp;
            yyn++;
            break;

          case '"':
            if (yyres)
              yyres[yyn] = '\0';
            return yyn;
          }
    do_not_strip_quotes: ;
    }

  if (! yyres)
    return yystrlen (yystr);

  return yystpcpy (yyres, yystr) - yyres;
}
# endif

/* Copy into *YYMSG, which is of size *YYMSG_ALLOC, an error message
   about the unexpected token YYTOKEN for the state stack whose top is
   YYSSP.

   Return 0 if *YYMSG was successfully written.  Return 1 if *YYMSG is
   not large enough to hold the message.  In that case, also set
   *YYMSG_ALLOC to the required number of bytes.  Return 2 if the
   required number of bytes is too large to store.  */
static int
yysyntax_error (YYSIZE_T *yymsg_alloc, char **yymsg,
                yytype_int16 *yyssp, int yytoken)
{
  YYSIZE_T yysize0 = yytnamerr (YY_NULLPTR, yytname[yytoken]);
  YYSIZE_T yysize = yysize0;
  enum { YYERROR_VERBOSE_ARGS_MAXIMUM = 5 };
  /* Internationalized format string. */
  const char *yyformat = YY_NULLPTR;
  /* Arguments of yyformat. */
  char const *yyarg[YYERROR_VERBOSE_ARGS_MAXIMUM];
  /* Number of reported tokens (one for the "unexpected", one per
     "expected"). */
  int yycount = 0;

  /* There are many possibilities here to consider:
     - If this state is a consistent state with a default action, then
       the only way this function was invoked is if the default action
       is an error action.  In that case, don't check for expected
       tokens because there are none.
     - The only way there can be no lookahead present (in yychar) is if
       this state is a consistent state with a default action.  Thus,
       detecting the absence of a lookahead is sufficient to determine
       that there is no unexpected or expected token to report.  In that
       case, just report a simple "syntax error".
     - Don't assume there isn't a lookahead just because this state is a
       consistent state with a default action.  There might have been a
       previous inconsistent state, consistent state with a non-default
       action, or user semantic action that manipulated yychar.
     - Of course, the expected token list depends on states to have
       correct lookahead information, and it depends on the parser not
       to perform extra reductions after fetching a lookahead from the
       scanner and before detecting a syntax error.  Thus, state merging
       (from LALR or IELR) and default reductions corrupt the expected
       token list.  However, the list is correct for canonical LR with
       one exception: it will still contain any token that will not be
       accepted due to an error action in a later state.
  */
  if (yytoken != YYEMPTY)
    {
      int yyn = yypact[*yyssp];
      yyarg[yycount++] = yytname[yytoken];
      if (!yypact_value_is_default (yyn))
        {
          /* Start YYX at -YYN if negative to avoid negative indexes in
             YYCHECK.  In other words, skip the first -YYN actions for
             this state because they are default actions.  */
          int yyxbegin = yyn < 0 ? -yyn : 0;
          /* Stay within bounds of both yycheck and yytname.  */
          int yychecklim = YYLAST - yyn + 1;
          int yyxend = yychecklim < YYNTOKENS ? yychecklim : YYNTOKENS;
          int yyx;

          for (yyx = yyxbegin; yyx < yyxend; ++yyx)
            if (yycheck[yyx + yyn] == yyx && yyx != YYTERROR
                && !yytable_value_is_error (yytable[yyx + yyn]))
              {
                if (yycount == YYERROR_VERBOSE_ARGS_MAXIMUM)
                  {
                    yycount = 1;
                    yysize = yysize0;
                    break;
                  }
                yyarg[yycount++] = yytname[yyx];
                {
                  YYSIZE_T yysize1 = yysize + yytnamerr (YY_NULLPTR, yytname[yyx]);
                  if (! (yysize <= yysize1
                         && yysize1 <= YYSTACK_ALLOC_MAXIMUM))
                    return 2;
                  yysize = yysize1;
                }
              }
        }
    }

  switch (yycount)
    {
# define YYCASE_(N, S)                      \
      case N:                               \
        yyformat = S;                       \
      break
      YYCASE_(0, YY_("syntax error"));
      YYCASE_(1, YY_("syntax error, unexpected %s"));
      YYCASE_(2, YY_("syntax error, unexpected %s, expecting %s"));
      YYCASE_(3, YY_("syntax error, unexpected %s, expecting %s or %s"));
      YYCASE_(4, YY_("syntax error, unexpected %s, expecting %s or %s or %s"));
      YYCASE_(5, YY_("syntax error, unexpected %s, expecting %s or %s or %s or %s"));
# undef YYCASE_
    }

  {
    YYSIZE_T yysize1 = yysize + yystrlen (yyformat);
    if (! (yysize <= yysize1 && yysize1 <= YYSTACK_ALLOC_MAXIMUM))
      return 2;
    yysize = yysize1;
  }

  if (*yymsg_alloc < yysize)
    {
      *yymsg_alloc = 2 * yysize;
      if (! (yysize <= *yymsg_alloc
             && *yymsg_alloc <= YYSTACK_ALLOC_MAXIMUM))
        *yymsg_alloc = YYSTACK_ALLOC_MAXIMUM;
      return 1;
    }

  /* Avoid sprintf, as that infringes on the user's name space.
     Don't have undefined behavior even if the translation
     produced a string with the wrong number of "%s"s.  */
  {
    char *yyp = *yymsg;
    int yyi = 0;
    while ((*yyp = *yyformat) != '\0')
      if (*yyp == '%' && yyformat[1] == 's' && yyi < yycount)
        {
          yyp += yytnamerr (yyp, yyarg[yyi++]);
          yyformat += 2;
        }
      else
        {
          yyp++;
          yyformat++;
        }
  }
  return 0;
}
#endif /* YYERROR_VERBOSE */

/*-----------------------------------------------.
| Release the memory associated to this symbol.  |
`-----------------------------------------------*/

static void
yydestruct (const char *yymsg, int yytype, YYSTYPE *yyvaluep)
{
  YYUSE (yyvaluep);
  if (!yymsg)
    yymsg = "Deleting";
  YY_SYMBOL_PRINT (yymsg, yytype, yyvaluep, yylocationp);

  YY_IGNORE_MAYBE_UNINITIALIZED_BEGIN
  YYUSE (yytype);
  YY_IGNORE_MAYBE_UNINITIALIZED_END
}




/* The lookahead symbol.  */
int yychar;

/* The semantic value of the lookahead symbol.  */
YYSTYPE yylval;
/* Number of syntax errors so far.  */
int yynerrs;


/*----------.
| yyparse.  |
`----------*/

int
yyparse (void)
{
    int yystate;
    /* Number of tokens to shift before error messages enabled.  */
    int yyerrstatus;

    /* The stacks and their tools:
       'yyss': related to states.
       'yyvs': related to semantic values.

       Refer to the stacks through separate pointers, to allow yyoverflow
       to reallocate them elsewhere.  */

    /* The state stack.  */
    yytype_int16 yyssa[YYINITDEPTH];
    yytype_int16 *yyss;
    yytype_int16 *yyssp;

    /* The semantic value stack.  */
    YYSTYPE yyvsa[YYINITDEPTH];
    YYSTYPE *yyvs;
    YYSTYPE *yyvsp;

    YYSIZE_T yystacksize;

  int yyn;
  int yyresult;
  /* Lookahead token as an internal (translated) token number.  */
  int yytoken = 0;
  /* The variables used to return semantic value and location from the
     action routines.  */
  YYSTYPE yyval;

#if YYERROR_VERBOSE
  /* Buffer for error messages, and its allocated size.  */
  char yymsgbuf[128];
  char *yymsg = yymsgbuf;
  YYSIZE_T yymsg_alloc = sizeof yymsgbuf;
#endif

#define YYPOPSTACK(N)   (yyvsp -= (N), yyssp -= (N))

  /* The number of symbols on the RHS of the reduced rule.
     Keep to zero when no symbol should be popped.  */
  int yylen = 0;

  yyssp = yyss = yyssa;
  yyvsp = yyvs = yyvsa;
  yystacksize = YYINITDEPTH;

  YYDPRINTF ((stderr, "Starting parse\n"));

  yystate = 0;
  yyerrstatus = 0;
  yynerrs = 0;
  yychar = YYEMPTY; /* Cause a token to be read.  */
  goto yysetstate;

/*------------------------------------------------------------.
| yynewstate -- Push a new state, which is found in yystate.  |
`------------------------------------------------------------*/
 yynewstate:
  /* In all cases, when you get here, the value and location stacks
     have just been pushed.  So pushing a state here evens the stacks.  */
  yyssp++;

 yysetstate:
  *yyssp = yystate;

  if (yyss + yystacksize - 1 <= yyssp)
    {
      /* Get the current used size of the three stacks, in elements.  */
      YYSIZE_T yysize = yyssp - yyss + 1;

#ifdef yyoverflow
      {
        /* Give user a chance to reallocate the stack.  Use copies of
           these so that the &'s don't force the real ones into
           memory.  */
        YYSTYPE *yyvs1 = yyvs;
        yytype_int16 *yyss1 = yyss;

        /* Each stack pointer address is followed by the size of the
           data in use in that stack, in bytes.  This used to be a
           conditional around just the two extra args, but that might
           be undefined if yyoverflow is a macro.  */
        yyoverflow (YY_("memory exhausted"),
                    &yyss1, yysize * sizeof (*yyssp),
                    &yyvs1, yysize * sizeof (*yyvsp),
                    &yystacksize);

        yyss = yyss1;
        yyvs = yyvs1;
      }
#else /* no yyoverflow */
# ifndef YYSTACK_RELOCATE
      goto yyexhaustedlab;
# else
      /* Extend the stack our own way.  */
      if (YYMAXDEPTH <= yystacksize)
        goto yyexhaustedlab;
      yystacksize *= 2;
      if (YYMAXDEPTH < yystacksize)
        yystacksize = YYMAXDEPTH;

      {
        yytype_int16 *yyss1 = yyss;
        union yyalloc *yyptr =
          (union yyalloc *) YYSTACK_ALLOC (YYSTACK_BYTES (yystacksize));
        if (! yyptr)
          goto yyexhaustedlab;
        YYSTACK_RELOCATE (yyss_alloc, yyss);
        YYSTACK_RELOCATE (yyvs_alloc, yyvs);
#  undef YYSTACK_RELOCATE
        if (yyss1 != yyssa)
          YYSTACK_FREE (yyss1);
      }
# endif
#endif /* no yyoverflow */

      yyssp = yyss + yysize - 1;
      yyvsp = yyvs + yysize - 1;

      YYDPRINTF ((stderr, "Stack size increased to %lu\n",
                  (unsigned long int) yystacksize));

      if (yyss + yystacksize - 1 <= yyssp)
        YYABORT;
    }

  YYDPRINTF ((stderr, "Entering state %d\n", yystate));

  if (yystate == YYFINAL)
    YYACCEPT;

  goto yybackup;

/*-----------.
| yybackup.  |
`-----------*/
yybackup:

  /* Do appropriate processing given the current state.  Read a
     lookahead token if we need one and don't already have one.  */

  /* First try to decide what to do without reference to lookahead token.  */
  yyn = yypact[yystate];
  if (yypact_value_is_default (yyn))
    goto yydefault;

  /* Not known => get a lookahead token if don't already have one.  */

  /* YYCHAR is either YYEMPTY or YYEOF or a valid lookahead symbol.  */
  if (yychar == YYEMPTY)
    {
      YYDPRINTF ((stderr, "Reading a token: "));
      yychar = yylex ();
    }

  if (yychar <= YYEOF)
    {
      yychar = yytoken = YYEOF;
      YYDPRINTF ((stderr, "Now at end of input.\n"));
    }
  else
    {
      yytoken = YYTRANSLATE (yychar);
      YY_SYMBOL_PRINT ("Next token is", yytoken, &yylval, &yylloc);
    }

  /* If the proper action on seeing token YYTOKEN is to reduce or to
     detect an error, take that action.  */
  yyn += yytoken;
  if (yyn < 0 || YYLAST < yyn || yycheck[yyn] != yytoken)
    goto yydefault;
  yyn = yytable[yyn];
  if (yyn <= 0)
    {
      if (yytable_value_is_error (yyn))
        goto yyerrlab;
      yyn = -yyn;
      goto yyreduce;
    }

  /* Count tokens shifted since error; after three, turn off error
     status.  */
  if (yyerrstatus)
    yyerrstatus--;

  /* Shift the lookahead token.  */
  YY_SYMBOL_PRINT ("Shifting", yytoken, &yylval, &yylloc);

  /* Discard the shifted token.  */
  yychar = YYEMPTY;

  yystate = yyn;
  YY_IGNORE_MAYBE_UNINITIALIZED_BEGIN
  *++yyvsp = yylval;
  YY_IGNORE_MAYBE_UNINITIALIZED_END

  goto yynewstate;


/*-----------------------------------------------------------.
| yydefault -- do the default action for the current state.  |
`-----------------------------------------------------------*/
yydefault:
  yyn = yydefact[yystate];
  if (yyn == 0)
    goto yyerrlab;
  goto yyreduce;


/*-----------------------------.
| yyreduce -- Do a reduction.  |
`-----------------------------*/
yyreduce:
  /* yyn is the number of a rule to reduce with.  */
  yylen = yyr2[yyn];

  /* If YYLEN is nonzero, implement the default value of the action:
     '$$ = $1'.

     Otherwise, the following line sets YYVAL to garbage.
     This behavior is undocumented and Bison
     users should not rely upon it.  Assigning to YYVAL
     unconditionally makes the parser a bit smaller, and it avoids a
     GCC warning that YYVAL may be used uninitialized.  */
  yyval = yyvsp[1-yylen];


  YY_REDUCE_PRINT (yyn);
  switch (yyn)
    {
        case 2:
#line 403 "PL.y" /* yacc.c:1646  */
    {
			  EXPR = *(yyvsp[-1].node);
			  delete (yyvsp[-1].node);
			  YYACCEPT;
			}
#line 3205 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 3:
#line 409 "PL.y" /* yacc.c:1646  */
    {
			  EXPR = *(yyvsp[-1].node);
			  delete (yyvsp[-1].node);
			  YYACCEPT;
			}
#line 3215 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 4:
#line 414 "PL.y" /* yacc.c:1646  */
    {
			  EXPR = *(yyvsp[-1].node);
			  delete (yyvsp[-1].node);
			  YYACCEPT;
			}
#line 3225 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 5:
#line 420 "PL.y" /* yacc.c:1646  */
    {
			  EXPR = CVC3::Expr();
			  YYACCEPT;
			}
#line 3234 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 6:
#line 425 "PL.y" /* yacc.c:1646  */
    {
			  TMP->done = true;
			  EXPR = CVC3::Expr();
			  YYACCEPT;
			}
#line 3244 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 7:
#line 432 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3250 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 8:
#line 433 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3256 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 9:
#line 434 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3262 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 10:
#line 435 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3268 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 11:
#line 436 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3274 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 12:
#line 437 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3280 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 13:
#line 438 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3286 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 14:
#line 439 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3292 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 15:
#line 440 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3298 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 16:
#line 441 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3304 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 17:
#line 442 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3310 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 18:
#line 443 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3316 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 19:
#line 444 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3322 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 20:
#line 445 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3328 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 21:
#line 446 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3334 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 22:
#line 447 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3340 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 23:
#line 448 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3346 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 24:
#line 449 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3352 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 25:
#line 450 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3358 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 26:
#line 451 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3364 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 27:
#line 452 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3370 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 28:
#line 453 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3376 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 29:
#line 454 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3382 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 30:
#line 455 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3388 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 31:
#line 459 "PL.y" /* yacc.c:1646  */
    {
                            (yyval.node) = new CVC3::Expr(VC->listExpr("_ARITH_VAR_ORDER", *(yyvsp[-1].vec)));
                            delete (yyvsp[-1].vec);
                        }
#line 3397 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 32:
#line 465 "PL.y" /* yacc.c:1646  */
    { 

                          (yyval.node) = new CVC3::Expr(VC->listExpr("_ASSERT", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3407 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 33:
#line 472 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_QUERY", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3416 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 34:
#line 476 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_CHECKSAT", *(yyvsp[0].node)));
                          delete (yyvsp[0].node);
                        }
#line 3425 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 35:
#line 480 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_CHECKSAT")));
                        }
#line 3433 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 36:
#line 483 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_CONTINUE")));
                        }
#line 3441 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 37:
#line 486 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_RESTART", *(yyvsp[0].node)));
                          delete (yyvsp[0].node);
                        }
#line 3450 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 38:
#line 491 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_DBG", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3459 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 39:
#line 495 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_DBG")));
		}
#line 3467 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 40:
#line 499 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_TRACE", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3476 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 41:
#line 503 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_UNTRACE", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3485 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 42:
#line 508 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_OPTION",  *(yyvsp[-1].node), *(yyvsp[0].node)));
			  delete (yyvsp[-1].node);
			  delete (yyvsp[0].node);
                        }
#line 3495 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 43:
#line 513 "PL.y" /* yacc.c:1646  */
    {
			  CVC3::Rational value= -(yyvsp[0].node)->getRational();
                          CVC3::Expr e = CVC3::Expr(VC->ratExpr(value.toString()));
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_OPTION",  *(yyvsp[-2].node), e));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
                        }
#line 3507 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 44:
#line 520 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_OPTION",  *(yyvsp[-1].node),  *(yyvsp[0].node)));
			  delete (yyvsp[-1].node);
			  delete (yyvsp[0].node);
                        }
#line 3517 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 45:
#line 526 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_HELP", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3526 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 46:
#line 530 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_HELP")));
		}
#line 3534 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 47:
#line 535 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_TRANSFORM", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3543 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 48:
#line 541 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_PRINT", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3552 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 49:
#line 545 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_PRINT", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3561 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 50:
#line 551 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_CALL", *(yyvsp[-1].node), *(yyvsp[0].node)));
			  delete (yyvsp[-1].node);
			  delete (yyvsp[0].node);
                        }
#line 3571 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 51:
#line 558 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_ECHO", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3580 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 52:
#line 562 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_ECHO")));
                        }
#line 3588 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 53:
#line 567 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_INCLUDE",  *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3597 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 54:
#line 573 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_DUMP_PROOF")));
                        }
#line 3605 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 55:
#line 576 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_DUMP_ASSUMPTIONS")));
                        }
#line 3613 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 56:
#line 579 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_DUMP_SIG")));
                        }
#line 3621 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 57:
#line 582 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_DUMP_TCC")));
                        }
#line 3629 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 58:
#line 585 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_DUMP_TCC_ASSUMPTIONS")));
                        }
#line 3637 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 59:
#line 588 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_DUMP_TCC_PROOF")));
                        }
#line 3645 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 60:
#line 591 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_DUMP_CLOSURE")));
                        }
#line 3653 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 61:
#line 594 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_DUMP_CLOSURE_PROOF")));
                        }
#line 3661 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 62:
#line 599 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_WHERE")));
               		}
#line 3669 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 63:
#line 602 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_ASSERTIONS")));
               		}
#line 3677 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 64:
#line 605 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_ASSUMPTIONS")));
               		}
#line 3685 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 65:
#line 608 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_COUNTEREXAMPLE")));
               		}
#line 3693 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 66:
#line 611 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_COUNTERMODEL")));
               		}
#line 3701 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 67:
#line 615 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_PUSH", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3710 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 68:
#line 619 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_PUSH")));
		        }
#line 3718 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 69:
#line 622 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_PUSH_SCOPE", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3727 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 70:
#line 626 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_PUSH_SCOPE")));
		        }
#line 3735 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 71:
#line 630 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_POP", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3744 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 72:
#line 634 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_POP")));
		        }
#line 3752 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 73:
#line 637 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_POP_SCOPE", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3761 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 74:
#line 641 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_POP_SCOPE")));
		        }
#line 3769 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 75:
#line 645 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_POPTO", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3778 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 76:
#line 649 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_POPTO")));
		        }
#line 3786 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 77:
#line 652 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_POPTO_SCOPE", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3795 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 78:
#line 656 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_POPTO_SCOPE")));
		        }
#line 3803 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 79:
#line 659 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_RESET")));
		        }
#line 3811 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 80:
#line 663 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_CONTEXT",  *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3820 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 81:
#line 667 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_CONTEXT")));
		        }
#line 3828 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 82:
#line 671 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_FORGET", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3837 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 83:
#line 676 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_GET_TYPE", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3846 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 84:
#line 681 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_CHECK_TYPE",*(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
                        }
#line 3856 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 85:
#line 687 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_GET_CHILD", *(yyvsp[-1].node), *(yyvsp[0].node)));
			  delete (yyvsp[-1].node);
			  delete (yyvsp[0].node);
                        }
#line 3866 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 86:
#line 693 "PL.y" /* yacc.c:1646  */
    { 
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_GET_CHILD", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
                        }
#line 3875 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 87:
#line 698 "PL.y" /* yacc.c:1646  */
    { 
                          std::vector<CVC3::Expr> tmp;
			  tmp.push_back(*(yyvsp[-9].node));
			  tmp.push_back(*(yyvsp[-7].node));
			  tmp.push_back(*(yyvsp[-5].node));
			  tmp.push_back(*(yyvsp[-3].node));
			  tmp.push_back(*(yyvsp[-1].node));
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_SUBSTITUTE", tmp));
			  delete (yyvsp[-9].node);
			  delete (yyvsp[-7].node);
			  delete (yyvsp[-5].node);
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 3894 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 88:
#line 714 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->idExpr(*(yyvsp[0].str)));
			  delete (yyvsp[0].str);
			}
#line 3903 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 89:
#line 720 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->stringExpr(*(yyvsp[0].str)));
			  delete (yyvsp[0].str);
			}
#line 3912 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 90:
#line 726 "PL.y" /* yacc.c:1646  */
    {
  			  (yyval.node) = new CVC3::Expr(VC->ratExpr((*(yyvsp[0].str))));
  			  delete (yyvsp[0].str);
			}
#line 3921 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 91:
#line 733 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr
			    (VC->listExpr("_BVCONST", VC->stringExpr(*(yyvsp[0].str))));
			  delete (yyvsp[0].str);
                        }
#line 3931 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 92:
#line 742 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr
			    (VC->listExpr("_BVCONST", VC->stringExpr(*(yyvsp[0].str)), VC->ratExpr(16)));
			  delete (yyvsp[0].str);
                        }
#line 3941 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 93:
#line 752 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3947 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 94:
#line 753 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3953 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 95:
#line 756 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3959 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 96:
#line 757 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3965 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 97:
#line 758 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3971 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 98:
#line 759 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3977 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 99:
#line 760 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3983 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 100:
#line 761 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3989 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 101:
#line 762 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 3995 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 102:
#line 763 "PL.y" /* yacc.c:1646  */
    {(yyval.node) = (yyvsp[0].node);}
#line 4001 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 103:
#line 764 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4007 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 104:
#line 765 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[-1].node); }
#line 4013 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 105:
#line 768 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4019 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 106:
#line 769 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4025 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 107:
#line 773 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(
                            VC->listExpr("_TYPEDEF",
                                         VC->listExpr("_DATATYPE", *(yyvsp[-1].vec))));
			  delete (yyvsp[-1].vec);
			}
#line 4036 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 108:
#line 782 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
			}
#line 4046 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 109:
#line 788 "PL.y" /* yacc.c:1646  */
    { 
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node)); 
                          (yyval.vec) = (yyvsp[-2].vec);
		          delete (yyvsp[0].node);
			}
#line 4056 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 110:
#line 796 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-2].node), VC->listExpr(*(yyvsp[0].vec))));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].vec);
                        }
#line 4066 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 111:
#line 804 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node); 
			}
#line 4076 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 112:
#line 810 "PL.y" /* yacc.c:1646  */
    { 
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node)); 
                          (yyval.vec) = (yyvsp[-2].vec); 
		          delete (yyvsp[0].node); 
			}
#line 4086 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 113:
#line 818 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[0].node)));
			  delete (yyvsp[0].node); 
			}
#line 4095 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 114:
#line 823 "PL.y" /* yacc.c:1646  */
    {
                          CVC3::Expr tmp = VC->listExpr(*(yyvsp[-1].vec));
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-3].node), tmp));
			  delete (yyvsp[-3].node); delete (yyvsp[-1].vec);
			}
#line 4105 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 115:
#line 831 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
			}
#line 4115 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 116:
#line 837 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
			  (yyval.vec) = (yyvsp[-2].vec); 
			  delete (yyvsp[0].node);
			}
#line 4125 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 117:
#line 845 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4135 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 118:
#line 853 "PL.y" /* yacc.c:1646  */
    {
                          // Old style functions
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_OLD_ARROW", *(yyvsp[-3].node), *(yyvsp[-1].node)));
                          delete (yyvsp[-3].node); delete (yyvsp[-1].node);
			}
#line 4145 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 119:
#line 859 "PL.y" /* yacc.c:1646  */
    {
			  std::vector<CVC3::Expr> temp;
			  temp.push_back(*(yyvsp[-2].node));
			  temp.push_back(*(yyvsp[0].node));

			  (yyval.node) = new CVC3::Expr(VC->listExpr("_ARROW", temp));
			  delete (yyvsp[-2].node); delete (yyvsp[0].node); 
			}
#line 4158 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 120:
#line 868 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-3].vec)->push_back(*(yyvsp[0].node));
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_ARROW", *(yyvsp[-3].vec)));
			  delete (yyvsp[-3].vec); delete (yyvsp[0].node); 
			}
#line 4168 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 121:
#line 876 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-1].vec)));
			  delete (yyvsp[-1].vec);
			}
#line 4177 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 122:
#line 883 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(VC->idExpr("_RECORD_TYPE"));
			  (yyval.vec)->push_back(*(yyvsp[0].node)); 
			  delete (yyvsp[0].node); 
                        }
#line 4188 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 123:
#line 890 "PL.y" /* yacc.c:1646  */
    { 
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node)); 
			  (yyval.vec) = (yyvsp[-2].vec); 
			  delete (yyvsp[0].node);
			}
#line 4198 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 124:
#line 898 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4208 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 125:
#line 906 "PL.y" /* yacc.c:1646  */
    {
			   (yyval.node) = new CVC3::Expr(VC->listExpr("_TUPLE_TYPE", *(yyvsp[-1].vec)));
			   delete (yyvsp[-1].vec);
			 }
#line 4217 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 126:
#line 913 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[-2].node));
                          (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[-2].node);
                          delete (yyvsp[0].node);
			}
#line 4229 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 127:
#line 921 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
			  (yyval.vec) = (yyvsp[-2].vec); 
			  delete (yyvsp[0].node);
			}
#line 4239 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 128:
#line 937 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_ARRAY", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4249 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 129:
#line 945 "PL.y" /* yacc.c:1646  */
    {
			   std::vector<CVC3::Expr>::iterator theIter = (yyvsp[-1].vec)->begin();
			   (yyvsp[-1].vec)->insert(theIter, VC->idExpr("_SCALARTYPE"));
			   (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-1].vec)));
			   delete (yyvsp[-1].vec);
			 }
#line 4260 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 130:
#line 954 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_SUBTYPE", *(yyvsp[-1].node)));
			  delete (yyvsp[-1].node);
			}
#line 4269 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 131:
#line 959 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_SUBTYPE", *(yyvsp[-3].node), *(yyvsp[-1].node)));
                          delete (yyvsp[-3].node);
                          delete (yyvsp[-1].node);
                        }
#line 4279 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 132:
#line 967 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->idExpr("_BOOLEAN"));
			}
#line 4287 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 135:
#line 975 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BITVECTOR", *(yyvsp[-1].node)));
			  delete (yyvsp[-1].node);
			}
#line 4296 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 136:
#line 982 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->idExpr("_REAL"));
			}
#line 4304 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 137:
#line 988 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->idExpr("_INT"));
                        }
#line 4312 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 138:
#line 993 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_SUBRANGE", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4322 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 139:
#line 1001 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_NEGINF")));
			}
#line 4330 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 140:
#line 1004 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4336 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 141:
#line 1005 "PL.y" /* yacc.c:1646  */
    {
                          CVC3::Rational value= -(yyvsp[0].node)->getRational();
			  (yyval.node) = new CVC3::Expr(VC->ratExpr(value.toString()));
			  delete (yyvsp[0].node);
                        }
#line 4346 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 142:
#line 1013 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr(VC->idExpr("_POSINF")));
			}
#line 4354 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 143:
#line 1016 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4360 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 144:
#line 1017 "PL.y" /* yacc.c:1646  */
    {
                          CVC3::Rational value= -(yyvsp[0].node)->getRational();
			  (yyval.node) = new CVC3::Expr(VC->ratExpr(value.toString()));
			  delete (yyvsp[0].node);
                        }
#line 4370 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 145:
#line 1026 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
			}
#line 4380 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 146:
#line 1032 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[0].vec)->push_back(*(yyvsp[-2].node));
			  (yyval.vec) = (yyvsp[0].vec); 
			  delete (yyvsp[-2].node);
			}
#line 4390 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 147:
#line 1040 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>((yyvsp[0].vec)->rbegin(),
							   (yyvsp[0].vec)->rend());
			  delete (yyvsp[0].vec);
			}
#line 4400 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 148:
#line 1049 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4406 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 149:
#line 1050 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4412 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 150:
#line 1051 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4418 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 151:
#line 1052 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4424 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 152:
#line 1054 "PL.y" /* yacc.c:1646  */
    {
			  std::vector<CVC3::Expr> tmp;
			  tmp.push_back(*(yyvsp[-3].node));
			  tmp.insert(tmp.end(), (yyvsp[-1].vec)->begin(), (yyvsp[-1].vec)->end());
			  (yyval.node) = new CVC3::Expr(VC->listExpr(tmp));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].vec);
			}
#line 4437 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 153:
#line 1063 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_SIMULATE", *(yyvsp[-1].vec)));
			  delete (yyvsp[-1].vec);
			}
#line 4446 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 154:
#line 1068 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_READ", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4456 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 155:
#line 1074 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_RECORD_SELECT", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4466 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 156:
#line 1080 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_TUPLE_SELECT", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4476 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 157:
#line 1086 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(CVC3::PLprocessUpdates(*(yyvsp[-2].node), *(yyvsp[0].vec)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].vec);
			}
#line 4486 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 158:
#line 1091 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4492 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 159:
#line 1092 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4498 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 160:
#line 1093 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4504 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 161:
#line 1095 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4510 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 162:
#line 1097 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4516 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 163:
#line 1098 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4522 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 164:
#line 1099 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 4528 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 165:
#line 1102 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->idExpr("_TRUE_EXPR"));
			}
#line 4536 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 166:
#line 1106 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->idExpr("_FALSE_EXPR"));
			}
#line 4544 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 167:
#line 1111 "PL.y" /* yacc.c:1646  */
    {
			  if ((yyvsp[0].node)->isRational())
			  {
			    CVC3::Rational value= -(yyvsp[0].node)->getRational();
			    (yyval.node)= new CVC3::Expr(VC->ratExpr(value.toString()));
			  }
			  else
			    (yyval.node) = new CVC3::Expr(VC->listExpr("_UMINUS", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
			}
#line 4559 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 168:
#line 1122 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_NOT", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
			}
#line 4568 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 169:
#line 1127 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_IS_INTEGER", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
			}
#line 4577 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 170:
#line 1132 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_TCC", *(yyvsp[-1].node)));
			  delete (yyvsp[-1].node);
			}
#line 4586 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 171:
#line 1137 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_EQ", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4596 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 172:
#line 1143 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_NEQ", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4606 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 173:
#line 1149 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_XOR", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4616 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 174:
#line 1155 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_OR", *(yyvsp[0].vec)));
			  delete (yyvsp[0].vec);
			}
#line 4625 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 175:
#line 1160 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_AND", *(yyvsp[0].vec)));
			  delete (yyvsp[0].vec);
			}
#line 4634 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 176:
#line 1165 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_IMPLIES", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4644 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 177:
#line 1171 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_IFF", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4654 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 178:
#line 1177 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_PLUS", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4664 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 179:
#line 1183 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_MINUS", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4674 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 180:
#line 1189 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_TRANS_CLOSURE",
							   *(yyvsp[-6].node), *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-6].node);
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4686 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 181:
#line 1197 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_MULT", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4696 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 182:
#line 1203 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_POW", *(yyvsp[0].node), *(yyvsp[-2].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4706 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 183:
#line 1209 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_DIVIDE", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4716 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 184:
#line 1227 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_GT", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4726 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 185:
#line 1233 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_GE", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4736 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 186:
#line 1239 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_FLOOR", *(yyvsp[-1].node)));
			  delete (yyvsp[-1].node);
			}
#line 4745 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 187:
#line 1244 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_LT", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4755 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 188:
#line 1250 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_LE", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4765 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 189:
#line 1256 "PL.y" /* yacc.c:1646  */
    {
			   (yyval.node) = (yyvsp[-1].node);
			 }
#line 4773 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 190:
#line 1260 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVNEG", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
			}
#line 4782 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 191:
#line 1265 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr
			    (VC->listExpr("_EXTRACT", VC->ratExpr(*(yyvsp[-3].str)),
                                          VC->ratExpr(*(yyvsp[-1].str)), *(yyvsp[-5].node)));
			  delete (yyvsp[-5].node);
			  delete (yyvsp[-3].str);
  			  delete (yyvsp[-1].str);
			}
#line 4795 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 192:
#line 1274 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVAND", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4805 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 193:
#line 1280 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVOR", *(yyvsp[-2].node), *(yyvsp[0].node))); 
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 4815 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 194:
#line 1286 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr
			    (VC->listExpr("_LEFTSHIFT", *(yyvsp[-2].node), VC->ratExpr(*(yyvsp[0].str))));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].str);
                        }
#line 4826 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 195:
#line 1293 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr
			    (VC->listExpr("_RIGHTSHIFT", *(yyvsp[-2].node), VC->ratExpr(*(yyvsp[0].str))));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].str);
                        }
#line 4837 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 196:
#line 1300 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVXOR", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4847 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 197:
#line 1306 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVNAND", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4857 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 198:
#line 1312 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVNOR", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4867 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 199:
#line 1318 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVCOMP", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4877 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 200:
#line 1324 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVXNOR", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4887 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 201:
#line 1330 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_SX",*(yyvsp[-3].node),VC->ratExpr(*(yyvsp[-1].str))));
			  delete (yyvsp[-3].node);
  			  delete (yyvsp[-1].str);
		        }
#line 4897 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 202:
#line 1336 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_BVZEROEXTEND",VC->ratExpr(*(yyvsp[-1].str)),*(yyvsp[-3].node)));
			  delete (yyvsp[-3].node);
  			  delete (yyvsp[-1].str);
                        }
#line 4907 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 203:
#line 1342 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_BVREPEAT",VC->ratExpr(*(yyvsp[-1].str)),*(yyvsp[-3].node)));
			  delete (yyvsp[-3].node);
  			  delete (yyvsp[-1].str);
                        }
#line 4917 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 204:
#line 1348 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_BVROTL",VC->ratExpr(*(yyvsp[-1].str)),*(yyvsp[-3].node)));
			  delete (yyvsp[-3].node);
  			  delete (yyvsp[-1].str);
                        }
#line 4927 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 205:
#line 1354 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_BVROTR",VC->ratExpr(*(yyvsp[-1].str)),*(yyvsp[-3].node)));
			  delete (yyvsp[-3].node);
  			  delete (yyvsp[-1].str);
                        }
#line 4937 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 206:
#line 1360 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVLT", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4947 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 207:
#line 1366 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVGT", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4957 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 208:
#line 1372 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVLE", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4967 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 209:
#line 1378 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVGE", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4977 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 210:
#line 1385 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVSLT", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4987 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 211:
#line 1391 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVSGT", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 4997 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 212:
#line 1397 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVSLE", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 5007 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 213:
#line 1403 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVSGE", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 5017 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 214:
#line 1409 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_CONCAT", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5027 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 215:
#line 1415 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr
			    (VC->listExpr("_INTTOBV", *(yyvsp[-5].node), VC->ratExpr(*(yyvsp[-3].str)),
					  VC->ratExpr(*(yyvsp[-1].str))));
			  delete (yyvsp[-5].node);
			  delete (yyvsp[-3].str);
			  delete (yyvsp[-1].str);
			}
#line 5040 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 216:
#line 1424 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVTOINT", *(yyvsp[-1].node)));
			  delete (yyvsp[-1].node);
                        }
#line 5049 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 217:
#line 1429 "PL.y" /* yacc.c:1646  */
    {
			  //FIXME: this function is not to be exposed
			  //to the user counterexamples containing
			  //this function can be translated into
			  //BV-EXTRACT and comparison with 0 or 1
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BOOLEXTRACT", *(yyvsp[-3].node),
                                              VC->ratExpr(*(yyvsp[-1].str))));
			  delete (yyvsp[-3].node);
  			  delete (yyvsp[-1].str);
                        }
#line 5064 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 218:
#line 1440 "PL.y" /* yacc.c:1646  */
    {
			  std::vector<CVC3::Expr> k;
			  k.push_back(VC->ratExpr(*(yyvsp[-3].str)));
			  k.insert(k.end(), (yyvsp[-1].vec)->begin(), (yyvsp[-1].vec)->end()); 
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_BVPLUS", k));
			  delete (yyvsp[-3].str);
			  delete (yyvsp[-1].vec);
                        }
#line 5077 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 219:
#line 1449 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr
                           (VC->listExpr("_BVSUB", VC->ratExpr(*(yyvsp[-5].str)), *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-5].str);
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5089 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 220:
#line 1457 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr
                           (VC->listExpr("_BVUDIV", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5100 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 221:
#line 1464 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr
                           (VC->listExpr("_BVSDIV", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5111 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 222:
#line 1471 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr
                           (VC->listExpr("_BVUREM", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5122 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 223:
#line 1478 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr
                           (VC->listExpr("_BVSREM", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5133 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 224:
#line 1485 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr
                           (VC->listExpr("_BVSMOD", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5144 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 225:
#line 1492 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr
                           (VC->listExpr("_BVSHL", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5155 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 226:
#line 1499 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr
                           (VC->listExpr("_BVASHR", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5166 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 227:
#line 1506 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr
                           (VC->listExpr("_BVLSHR", *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5177 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 228:
#line 1513 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =  new CVC3::Expr(VC->listExpr("_BVUMINUS", *(yyvsp[-1].node)));
			  delete (yyvsp[-1].node);
                        }
#line 5186 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 229:
#line 1518 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr
			    (VC->listExpr("_BVMULT", VC->ratExpr(*(yyvsp[-5].str)), *(yyvsp[-3].node), *(yyvsp[-1].node)));
			  delete (yyvsp[-5].str);
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
                        }
#line 5198 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 230:
#line 1526 "PL.y" /* yacc.c:1646  */
    {
                          (yyval.node) = new CVC3::Expr(VC->listExpr("_DISTINCT", *(yyvsp[-1].vec)));
                          delete (yyvsp[-1].vec);
                        }
#line 5207 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 231:
#line 1532 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
                          (yyval.vec) = (yyvsp[-2].vec);
			  delete (yyvsp[0].node);
			}
#line 5217 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 232:
#line 1539 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[-2].node));
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
                        }
#line 5229 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 233:
#line 1550 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
                          (yyval.vec) = (yyvsp[-2].vec);
			  delete (yyvsp[0].node);
			}
#line 5239 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 234:
#line 1557 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[-2].node));
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
                        }
#line 5251 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 235:
#line 1566 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[0].vec)->push_back(VC->listExpr(*(yyvsp[-3].node), *(yyvsp[-1].node)));
			  (yyvsp[0].vec)->push_back(VC->idExpr("_COND"));
			  /* at this point, the list for ElseRest is 
			     in reverse order from what it should be. */
			  std::vector<CVC3::Expr> tmp;
			  tmp.insert(tmp.end(), (yyvsp[0].vec)->rbegin(), (yyvsp[0].vec)->rend());
			  (yyval.node) = new CVC3::Expr(VC->listExpr(tmp));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			  delete (yyvsp[0].vec);
			}
#line 5268 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 236:
#line 1581 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
	                  (yyval.vec)->push_back(VC->listExpr("_ELSE",*(yyvsp[-1].node)));
			  delete (yyvsp[-1].node);
			}
#line 5278 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 237:
#line 1587 "PL.y" /* yacc.c:1646  */
    {
			  /* NOTE that we're getting the list built
			     up in the reverse order here.  We'll fix
			     things when we produce a Conditional. */
			  (yyvsp[0].vec)->push_back(VC->listExpr(*(yyvsp[-3].node), *(yyvsp[-1].node)));
			  (yyval.vec) = (yyvsp[0].vec);
			  delete (yyvsp[-3].node);
			  delete (yyvsp[-1].node);
			}
#line 5292 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 238:
#line 1599 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
			}
#line 5302 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 239:
#line 1605 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
			  (yyval.vec) = (yyvsp[-2].vec); 
			  delete (yyvsp[0].node);
			}
#line 5312 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 240:
#line 1614 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = (yyvsp[-2].vec);
			}
#line 5320 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 241:
#line 1619 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(VC->listExpr(*(yyvsp[0].vec)));
			  delete (yyvsp[0].vec);
			}
#line 5330 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 242:
#line 1625 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-1].vec)->push_back(VC->listExpr(*(yyvsp[0].vec)));
			  (yyval.vec) = (yyvsp[-1].vec); 
			  delete (yyvsp[0].vec);
			}
#line 5340 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 243:
#line 1633 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5350 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 244:
#line 1641 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
			}
#line 5360 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 245:
#line 1647 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
			}
#line 5369 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 246:
#line 1654 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[0].vec)));
			  delete (yyvsp[0].vec);
			}
#line 5378 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 247:
#line 1661 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(VC->listExpr("_READ", *(yyvsp[-1].node)));
			  delete (yyvsp[-1].node);
			}
#line 5388 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 248:
#line 1667 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
			}
#line 5398 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 249:
#line 1673 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(VC->listExpr("_RECORD_SELECT", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
			}
#line 5408 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 250:
#line 1679 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(VC->listExpr("_TUPLE_SELECT", *(yyvsp[0].node)));
			  delete (yyvsp[0].node);
			}
#line 5418 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 251:
#line 1685 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-3].vec)->push_back(VC->listExpr("_READ", *(yyvsp[-1].node)));
			  (yyval.vec) = (yyvsp[-3].vec);
			  delete (yyvsp[-1].node);
			}
#line 5428 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 252:
#line 1691 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-1].vec)->push_back(*(yyvsp[0].node));
			  (yyval.vec) = (yyvsp[-1].vec);
			  delete (yyvsp[0].node);
			}
#line 5438 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 253:
#line 1697 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->push_back(VC->listExpr("_RECORD_SELECT",*(yyvsp[0].node)));
			  (yyval.vec) = (yyvsp[-2].vec);
			  delete (yyvsp[0].node);
			}
#line 5448 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 254:
#line 1703 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->push_back(VC->listExpr("_TUPLE_SELECT", *(yyvsp[0].node)));
			  (yyval.vec) = (yyvsp[-2].vec);
			  delete (yyvsp[0].node);
			}
#line 5458 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 255:
#line 1712 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_LAMBDA", 
			                                   VC->listExpr(*(yyvsp[-3].vec)), (*(yyvsp[0].node))));
			  delete (yyvsp[-3].vec);
			  delete (yyvsp[0].node);
			}
#line 5469 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 256:
#line 1721 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_FORALL", 
			                                   VC->listExpr(*(yyvsp[-3].vec)), *(yyvsp[0].node)));
			  delete (yyvsp[-3].vec);
			  delete (yyvsp[0].node);
			}
#line 5480 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 257:
#line 1729 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_FORALL", 
			                                   VC->listExpr(*(yyvsp[-4].vec)), *(yyvsp[0].node),
							   VC->listExpr(*(yyvsp[-1].vec))));
			  delete (yyvsp[-4].vec);
			  delete (yyvsp[-1].vec);
			  delete (yyvsp[0].node);
			}
#line 5493 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 258:
#line 1739 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_EXISTS", 
			                                   VC->listExpr(*(yyvsp[-3].vec)), (*(yyvsp[0].node))));
			  delete (yyvsp[-3].vec);
			  delete (yyvsp[0].node);
			}
#line 5504 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 259:
#line 1748 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_EXISTS", 
			                                   VC->listExpr(*(yyvsp[-4].vec)), *(yyvsp[0].node),
							   VC->listExpr(*(yyvsp[-1].vec))));
			  delete (yyvsp[-4].vec);
			  delete (yyvsp[-1].vec);
			  delete (yyvsp[0].node);
			}
#line 5517 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 260:
#line 1759 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr
			    (VC->listExpr("_ARRAY_LITERAL", *(yyvsp[-3].node), *(yyvsp[0].node)));
			  delete (yyvsp[-3].node);
			  delete (yyvsp[0].node);
			}
#line 5528 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 261:
#line 1767 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-1].vec)->insert((yyvsp[-1].vec)->begin(), VC->idExpr("_RECORD"));
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-1].vec)));
			  delete (yyvsp[-1].vec);
			}
#line 5538 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 262:
#line 1775 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node)); 
			  delete (yyvsp[0].node);
			}
#line 5548 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 263:
#line 1781 "PL.y" /* yacc.c:1646  */
    { 
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
			  (yyval.vec) = (yyvsp[-2].vec); 
			  delete (yyvsp[0].node);
			}
#line 5558 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 264:
#line 1789 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5568 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 265:
#line 1797 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-3].vec)->push_back(*(yyvsp[-1].node));
			  (yyvsp[-3].vec)->insert((yyvsp[-3].vec)->begin(),VC->idExpr("_TUPLE"));
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-3].vec)));
			  delete (yyvsp[-3].vec);
			  delete (yyvsp[-1].node);
			}
#line 5580 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 266:
#line 1807 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[0].vec)));
			  delete (yyvsp[0].vec);
			}
#line 5589 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 267:
#line 1812 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
                        }
#line 5599 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 268:
#line 1818 "PL.y" /* yacc.c:1646  */
    { 
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
			  (yyval.vec) = (yyvsp[-2].vec); 
			  delete (yyvsp[0].node);
			}
#line 5609 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 269:
#line 1826 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-2].node),*(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5619 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 270:
#line 1832 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-4].node),*(yyvsp[0].node)));
			  delete (yyvsp[-4].node);
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5630 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 271:
#line 1841 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_LET", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5640 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 272:
#line 1849 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[0].vec)));
			  delete (yyvsp[0].vec);
			}
#line 5649 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 273:
#line 1854 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
                        }
#line 5659 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 274:
#line 1860 "PL.y" /* yacc.c:1646  */
    { 
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
			  (yyval.vec) = (yyvsp[-2].vec); 
			  delete (yyvsp[0].node);
			}
#line 5669 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 275:
#line 1868 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5679 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 276:
#line 1874 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-4].node),*(yyvsp[0].node)));
			  delete (yyvsp[-4].node);
			  delete (yyvsp[0].node);
			}
#line 5689 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 277:
#line 1882 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_LET", *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5699 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 278:
#line 1890 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[-2].vec)));
			  delete (yyvsp[-2].vec);
			
			}
#line 5711 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 279:
#line 1900 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.vec) = new std::vector<CVC3::Expr>;
			  (yyval.vec)->push_back(*(yyvsp[0].node));
			  delete (yyvsp[0].node);
			}
#line 5721 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 280:
#line 1906 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->push_back(*(yyvsp[0].node));
			  (yyval.vec) = (yyvsp[-2].vec); 
			  delete (yyvsp[0].node);
			}
#line 5731 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 281:
#line 1914 "PL.y" /* yacc.c:1646  */
    { 
			  (yyval.node) = new CVC3::Expr(VC->listExpr(*(yyvsp[0].vec)));
			  delete (yyvsp[0].vec);
			}
#line 5740 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 282:
#line 1921 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_CONST",
							   *(yyvsp[-4].node), *(yyvsp[-2].node), *(yyvsp[0].node)));
			  delete (yyvsp[-4].node);
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5752 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 283:
#line 1929 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =new CVC3::Expr
			    (VC->listExpr("_CONST", VC->listExpr(*(yyvsp[-2].node)), *(yyvsp[0].node)));
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5763 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 284:
#line 1936 "PL.y" /* yacc.c:1646  */
    {
			  std::vector<CVC3::Expr> tmp;
			  tmp.push_back(VC->idExpr("_DEFUN"));
			  tmp.push_back(*(yyvsp[-7].node));
			  tmp.push_back(*(yyvsp[-5].node));
			  tmp.push_back(*(yyvsp[-2].node));
			  tmp.push_back(*(yyvsp[0].node));
			  (yyval.node) = new CVC3::Expr(VC->listExpr(tmp));
			  delete (yyvsp[-7].node);
			  delete (yyvsp[-5].node);
			  delete (yyvsp[-2].node);
			  delete (yyvsp[0].node);
			}
#line 5781 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 285:
#line 1950 "PL.y" /* yacc.c:1646  */
    {
			  std::vector<CVC3::Expr> tmp;
			  tmp.push_back(VC->idExpr("_DEFUN"));
			  tmp.push_back(*(yyvsp[-5].node));
			  tmp.push_back(*(yyvsp[-3].node));
			  tmp.push_back(*(yyvsp[0].node));
			  (yyval.node) = new CVC3::Expr(VC->listExpr(tmp));
			  delete (yyvsp[-5].node);
			  delete (yyvsp[-3].node);
			  delete (yyvsp[0].node);
			}
#line 5797 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 286:
#line 1962 "PL.y" /* yacc.c:1646  */
    {
			  (yyvsp[-2].vec)->insert((yyvsp[-2].vec)->begin(), *(yyvsp[-4].node));
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_CONST", VC->listExpr(*(yyvsp[-2].vec)), *(yyvsp[0].node)));
			  delete (yyvsp[-4].node);
			  delete (yyvsp[-2].vec);
			  delete (yyvsp[0].node);
			}
#line 5809 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 287:
#line 1971 "PL.y" /* yacc.c:1646  */
    { (yyval.node) = (yyvsp[0].node); }
#line 5815 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 288:
#line 1973 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_TYPEDEF", *(yyvsp[-4].node), *(yyvsp[0].node)));
			  delete (yyvsp[-4].node);
			  delete (yyvsp[0].node);
			}
#line 5825 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 289:
#line 1979 "PL.y" /* yacc.c:1646  */
    {
			  (yyval.node) =new CVC3::Expr(VC->listExpr("_TYPE", *(yyvsp[-2].node)));
			  delete (yyvsp[-2].node);
			}
#line 5834 "parsePL.cpp" /* yacc.c:1646  */
    break;

  case 290:
#line 1984 "PL.y" /* yacc.c:1646  */
    {
			  std::vector<CVC3::Expr> tmp;
			  tmp.push_back(*(yyvsp[-4].node));
			  tmp.insert(tmp.end(), (yyvsp[-2].vec)->begin(), (yyvsp[-2].vec)->end());
			  (yyval.node) = new CVC3::Expr(VC->listExpr("_TYPE", tmp));
			  delete (yyvsp[-4].node);
			  delete (yyvsp[-2].vec);
			}
#line 5847 "parsePL.cpp" /* yacc.c:1646  */
    break;


#line 5851 "parsePL.cpp" /* yacc.c:1646  */
      default: break;
    }
  /* User semantic actions sometimes alter yychar, and that requires
     that yytoken be updated with the new translation.  We take the
     approach of translating immediately before every use of yytoken.
     One alternative is translating here after every semantic action,
     but that translation would be missed if the semantic action invokes
     YYABORT, YYACCEPT, or YYERROR immediately after altering yychar or
     if it invokes YYBACKUP.  In the case of YYABORT or YYACCEPT, an
     incorrect destructor might then be invoked immediately.  In the
     case of YYERROR or YYBACKUP, subsequent parser actions might lead
     to an incorrect destructor call or verbose syntax error message
     before the lookahead is translated.  */
  YY_SYMBOL_PRINT ("-> $$ =", yyr1[yyn], &yyval, &yyloc);

  YYPOPSTACK (yylen);
  yylen = 0;
  YY_STACK_PRINT (yyss, yyssp);

  *++yyvsp = yyval;

  /* Now 'shift' the result of the reduction.  Determine what state
     that goes to, based on the state we popped back to and the rule
     number reduced by.  */

  yyn = yyr1[yyn];

  yystate = yypgoto[yyn - YYNTOKENS] + *yyssp;
  if (0 <= yystate && yystate <= YYLAST && yycheck[yystate] == *yyssp)
    yystate = yytable[yystate];
  else
    yystate = yydefgoto[yyn - YYNTOKENS];

  goto yynewstate;


/*--------------------------------------.
| yyerrlab -- here on detecting error.  |
`--------------------------------------*/
yyerrlab:
  /* Make sure we have latest lookahead translation.  See comments at
     user semantic actions for why this is necessary.  */
  yytoken = yychar == YYEMPTY ? YYEMPTY : YYTRANSLATE (yychar);

  /* If not already recovering from an error, report this error.  */
  if (!yyerrstatus)
    {
      ++yynerrs;
#if ! YYERROR_VERBOSE
      yyerror (YY_("syntax error"));
#else
# define YYSYNTAX_ERROR yysyntax_error (&yymsg_alloc, &yymsg, \
                                        yyssp, yytoken)
      {
        char const *yymsgp = YY_("syntax error");
        int yysyntax_error_status;
        yysyntax_error_status = YYSYNTAX_ERROR;
        if (yysyntax_error_status == 0)
          yymsgp = yymsg;
        else if (yysyntax_error_status == 1)
          {
            if (yymsg != yymsgbuf)
              YYSTACK_FREE (yymsg);
            yymsg = (char *) YYSTACK_ALLOC (yymsg_alloc);
            if (!yymsg)
              {
                yymsg = yymsgbuf;
                yymsg_alloc = sizeof yymsgbuf;
                yysyntax_error_status = 2;
              }
            else
              {
                yysyntax_error_status = YYSYNTAX_ERROR;
                yymsgp = yymsg;
              }
          }
        yyerror (yymsgp);
        if (yysyntax_error_status == 2)
          goto yyexhaustedlab;
      }
# undef YYSYNTAX_ERROR
#endif
    }



  if (yyerrstatus == 3)
    {
      /* If just tried and failed to reuse lookahead token after an
         error, discard it.  */

      if (yychar <= YYEOF)
        {
          /* Return failure if at end of input.  */
          if (yychar == YYEOF)
            YYABORT;
        }
      else
        {
          yydestruct ("Error: discarding",
                      yytoken, &yylval);
          yychar = YYEMPTY;
        }
    }

  /* Else will try to reuse lookahead token after shifting the error
     token.  */
  goto yyerrlab1;


/*---------------------------------------------------.
| yyerrorlab -- error raised explicitly by YYERROR.  |
`---------------------------------------------------*/
yyerrorlab:

  /* Pacify compilers like GCC when the user code never invokes
     YYERROR and the label yyerrorlab therefore never appears in user
     code.  */
  if (/*CONSTCOND*/ 0)
     goto yyerrorlab;

  /* Do not reclaim the symbols of the rule whose action triggered
     this YYERROR.  */
  YYPOPSTACK (yylen);
  yylen = 0;
  YY_STACK_PRINT (yyss, yyssp);
  yystate = *yyssp;
  goto yyerrlab1;


/*-------------------------------------------------------------.
| yyerrlab1 -- common code for both syntax error and YYERROR.  |
`-------------------------------------------------------------*/
yyerrlab1:
  yyerrstatus = 3;      /* Each real token shifted decrements this.  */

  for (;;)
    {
      yyn = yypact[yystate];
      if (!yypact_value_is_default (yyn))
        {
          yyn += YYTERROR;
          if (0 <= yyn && yyn <= YYLAST && yycheck[yyn] == YYTERROR)
            {
              yyn = yytable[yyn];
              if (0 < yyn)
                break;
            }
        }

      /* Pop the current state because it cannot handle the error token.  */
      if (yyssp == yyss)
        YYABORT;


      yydestruct ("Error: popping",
                  yystos[yystate], yyvsp);
      YYPOPSTACK (1);
      yystate = *yyssp;
      YY_STACK_PRINT (yyss, yyssp);
    }

  YY_IGNORE_MAYBE_UNINITIALIZED_BEGIN
  *++yyvsp = yylval;
  YY_IGNORE_MAYBE_UNINITIALIZED_END


  /* Shift the error token.  */
  YY_SYMBOL_PRINT ("Shifting", yystos[yyn], yyvsp, yylsp);

  yystate = yyn;
  goto yynewstate;


/*-------------------------------------.
| yyacceptlab -- YYACCEPT comes here.  |
`-------------------------------------*/
yyacceptlab:
  yyresult = 0;
  goto yyreturn;

/*-----------------------------------.
| yyabortlab -- YYABORT comes here.  |
`-----------------------------------*/
yyabortlab:
  yyresult = 1;
  goto yyreturn;

#if !defined yyoverflow || YYERROR_VERBOSE
/*-------------------------------------------------.
| yyexhaustedlab -- memory exhaustion comes here.  |
`-------------------------------------------------*/
yyexhaustedlab:
  yyerror (YY_("memory exhausted"));
  yyresult = 2;
  /* Fall through.  */
#endif

yyreturn:
  if (yychar != YYEMPTY)
    {
      /* Make sure we have latest lookahead translation.  See comments at
         user semantic actions for why this is necessary.  */
      yytoken = YYTRANSLATE (yychar);
      yydestruct ("Cleanup: discarding lookahead",
                  yytoken, &yylval);
    }
  /* Do not reclaim the symbols of the rule whose action triggered
     this YYABORT or YYACCEPT.  */
  YYPOPSTACK (yylen);
  YY_STACK_PRINT (yyss, yyssp);
  while (yyssp != yyss)
    {
      yydestruct ("Cleanup: popping",
                  yystos[*yyssp], yyvsp);
      YYPOPSTACK (1);
    }
#ifndef yyoverflow
  if (yyss != yyssa)
    YYSTACK_FREE (yyss);
#endif
#if YYERROR_VERBOSE
  if (yymsg != yymsgbuf)
    YYSTACK_FREE (yymsg);
#endif
  return yyresult;
}
#line 1993 "PL.y" /* yacc.c:1906  */

